/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.cpp
    \brief    prediction class
*/

#include <memory.h>
#include "TComPrediction.h"
#if SHVC_COMPLEXITY_ASSESSMENT
namespace {

/*
 * \brief Derive an AU-aligned address of a given address in a rounding-down manner.
 *
 * \param iAddr
 * \param iAccessUnit
 *
 * \return AU-aligned address
 */
__inline Int iAlignmentRoundDown( Int iAddr, Int iAccessUnit )
{
  Int iAlignedAddr = iAccessUnit * (iAddr/iAccessUnit);
  if ( iAddr < 0 )
  {
    iAlignedAddr -= (iAddr%iAccessUnit) ? iAccessUnit : 0;
  }
  return iAlignedAddr;
}

/*
 * \brief Derive an AU-aligned address of a given address in a rounding-off manner.
 *
 * \param iAddr
 * \param iAccessUnit
 *
 * \return AU-aligned address
 */
__inline Int iAlignmentRoundUp( Int iAddr, Int iAccessUnit )
{
  return iAccessUnit * (iAddr/iAccessUnit) + ((iAddr%iAccessUnit) ? iAccessUnit : 0);
}

} // end of nameless namespace
#endif //SHVC_COMPLEXITY_ASSESSMENT
//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Constructor / destructor / initialize
// ====================================================================================================================

TComPrediction::TComPrediction()
: m_pLumaRecBuffer(0)
, m_iLumaRecStride(0)
{
  m_piYuvExt = NULL;
#if SHVC_COMPLEXITY_ASSESSMENT
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    m_apcMemAccessCalc[i] = 0;
    m_apcMemAccessCalcBlock[i] = 0; 
	m_apcMemAccessCalcPic[i] = 0; 
#if ILR_TM_ENH_COMPLEXITY
  m_apcMemAccessCalcPic_enhILR[i] = 0;
#endif
  }
 g_d_num_pel_predicted_with_interpolation =0; g_d_num_adds=0; g_d_num_mults=0; 
 g_d_num_adds_pic=0; g_d_num_mults_pic=0; 
 g_d_num_adds_block=0; g_d_num_mults_block=0; 
 pic_level_calc_done=0;
#endif//SHVC_COMPLEXITY_ASSESSMENT
#if ILR_TM_ENH_COMPLEXITY
 g_d_num_adds_pic_enhILR=0; g_d_num_mults_pic_enhILR=0; 
#endif
#if ILR_TM_ENH
  memset(m_pEnhPred, 0, sizeof(m_pEnhPred));
#endif

}

TComPrediction::~TComPrediction()
{
  
  delete[] m_piYuvExt;

  m_acYuvPred[0].destroy();
  m_acYuvPred[1].destroy();

  m_cYuvPredTemp.destroy();
#if SHVC_COMPLEXITY_ASSESSMENT
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if ( m_apcMemAccessCalc[i] )
    {
      if ( m_apcMemAccessCalc[i]->getCache() )
      {
        delete m_apcMemAccessCalc[i]->getCache();
      }
      delete m_apcMemAccessCalc[i];
    }
    if ( m_apcMemAccessCalcPic[i] )
    {
      if ( m_apcMemAccessCalcPic[i]->getCache() )
      {
        delete m_apcMemAccessCalcPic[i]->getCache();
      }
      delete m_apcMemAccessCalcPic[i];
    }
    if ( m_apcMemAccessCalcBlock[i] )
    {
      if ( m_apcMemAccessCalcBlock[i]->getCache() )
      {
        delete m_apcMemAccessCalcBlock[i]->getCache();
      }
      delete m_apcMemAccessCalcBlock[i];
    }
#if ILR_TM_ENH_COMPLEXITY
    if ( m_apcMemAccessCalcPic_enhILR[i] )
    {
      if ( m_apcMemAccessCalcPic_enhILR[i]->getCache() )
      {
        delete m_apcMemAccessCalcPic_enhILR[i]->getCache();
      }
      delete m_apcMemAccessCalcPic_enhILR[i];
    }
#endif
  }
#endif //SHVC_COMPLEXITY_ASSESSMENT

  if( m_pLumaRecBuffer )
  {
    delete [] m_pLumaRecBuffer;
  }
  
  Int i, j;
  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
    {
      m_filteredBlock[i][j].destroy();
    }
    m_filteredBlockTmp[i].destroy();
  }
#if ILR_TM_ENH 
  if(m_pEnhPred[0])
  {
    m_pEnhPred[0]->destroy();
    delete m_pEnhPred[0];
    m_pEnhPred[0] = NULL;
  }
  if(m_pEnhPred[1])
  {
    m_pEnhPred[1]->destroy();
    delete m_pEnhPred[1];
    m_pEnhPred[1] = NULL;
  }
#endif
}
#if ILR_TM_ENH 
Void TComPrediction::initTempBuff(Bool bEnhPred)
#else
Void TComPrediction::initTempBuff()
#endif
{
  if( m_piYuvExt == NULL )
  {
    Int extWidth  = g_uiMaxCUWidth + 16; 
    Int extHeight = g_uiMaxCUHeight + 1;
    Int i, j;
    for (i = 0; i < 4; i++)
    {
      m_filteredBlockTmp[i].create(extWidth, extHeight + 7);
      for (j = 0; j < 4; j++)
      {
        m_filteredBlock[i][j].create(extWidth, extHeight);
      }
    }
    m_iYuvExtHeight  = ((g_uiMaxCUHeight + 2) << 4);
    m_iYuvExtStride = ((g_uiMaxCUWidth  + 8) << 4);
    m_piYuvExt = new Int[ m_iYuvExtStride * m_iYuvExtHeight ];

    // new structure
    m_acYuvPred[0] .create( g_uiMaxCUWidth, g_uiMaxCUHeight );
    m_acYuvPred[1] .create( g_uiMaxCUWidth, g_uiMaxCUHeight );

    m_cYuvPredTemp.create( g_uiMaxCUWidth, g_uiMaxCUHeight );

#if ILR_TM_ENH 
  if(bEnhPred)
  {
    m_pEnhPred[0] = new TComYuv;    
    m_pEnhPred[0]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
#if (!ILR_TM_ENH_MCD)
    m_pEnhPred[1] = new TComYuv;
    m_pEnhPred[1]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
#endif
  }
#endif

  }

  if (m_iLumaRecStride != (g_uiMaxCUWidth>>1) + 1)
  {
    m_iLumaRecStride =  (g_uiMaxCUWidth>>1) + 1;
    if (!m_pLumaRecBuffer)
    {
      m_pLumaRecBuffer = new Pel[ m_iLumaRecStride * m_iLumaRecStride ];
    }
  }

  Int shift = g_uiBitDepth + g_uiBitIncrement + 4;

  for( Int i = 32; i < 64; i++ )
  {
    m_uiaShift[i-32] = ( ( 1 << shift ) + i/2 ) / i;
  }
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

// Function for calculating DC value of the reference samples used in Intra prediction
Pel TComPrediction::predIntraGetPredValDC( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft )
{
  Int iInd, iSum = 0;
  Pel pDcVal;

  if (bAbove)
  {
    for (iInd = 0;iInd < iWidth;iInd++)
    {
      iSum += pSrc[iInd-iSrcStride];
    }
  }
  if (bLeft)
  {
    for (iInd = 0;iInd < iHeight;iInd++)
    {
      iSum += pSrc[iInd*iSrcStride-1];
    }
  }

  if (bAbove && bLeft)
  {
    pDcVal = (iSum + iWidth) / (iWidth + iHeight);
  }
  else if (bAbove)
  {
    pDcVal = (iSum + iWidth/2) / iWidth;
  }
  else if (bLeft)
  {
    pDcVal = (iSum + iHeight/2) / iHeight;
  }
  else
  {
    pDcVal = pSrc[-1]; // Default DC value already calculated and placed in the prediction array if no neighbors are available
  }
  
  return pDcVal;
}

// Function for deriving the angular Intra predictions

/** Function for deriving the simplified angular intra predictions.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 * \param dirMode the intra prediction mode index
 * \param blkAboveAvailable boolean indication if the block above is available
 * \param blkLeftAvailable boolean indication if the block to the left is available
 *
 * This function derives the prediction samples for the angular mode based on the prediction direction indicated by
 * the prediction mode index. The prediction direction is given by the displacement of the bottom row of the block and
 * the reference row above the block in the case of vertical prediction or displacement of the rightmost column
 * of the block and reference column left from the block in the case of the horizontal prediction. The displacement
 * is signalled at 1/32 pixel accuracy. When projection of the predicted pixel falls inbetween reference samples,
 * the predicted value for the pixel is linearly interpolated from the reference samples. All reference samples are taken
 * from the extended main reference.
 */
Void TComPrediction::xPredIntraAng( Int* pSrc, Int srcStride, Pel*& rpDst, Int dstStride, UInt width, UInt height, UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable, Bool bFilter )
{
  Int k,l;
  Int blkSize        = width;
  Pel* pDst          = rpDst;

  // Map the mode index to main prediction direction and angle
  assert( dirMode > 0 ); //no planar
  Bool modeDC        = dirMode < 2;
  Bool modeHor       = !modeDC && (dirMode < 18);
  Bool modeVer       = !modeDC && !modeHor;
  Int intraPredAngle = modeVer ? (Int)dirMode - VER_IDX : modeHor ? -((Int)dirMode - HOR_IDX) : 0;
  Int absAng         = abs(intraPredAngle);
  Int signAng        = intraPredAngle < 0 ? -1 : 1;

  // Set bitshifts and scale the angle parameter to block size
  Int angTable[9]    = {0,    2,    5,   9,  13,  17,  21,  26,  32};
  Int invAngTable[9] = {0, 4096, 1638, 910, 630, 482, 390, 315, 256}; // (256 * 32) / Angle
  Int invAngle       = invAngTable[absAng];
  absAng             = angTable[absAng];
  intraPredAngle     = signAng * absAng;

  // Do the DC prediction
  if (modeDC)
  {
    Pel dcval = predIntraGetPredValDC(pSrc, srcStride, width, height, blkAboveAvailable, blkLeftAvailable);

    for (k=0;k<blkSize;k++)
    {
      for (l=0;l<blkSize;l++)
      {
        pDst[k*dstStride+l] = dcval;
      }
    }
  }

  // Do angular predictions
  else
  {
    Pel* refMain;
    Pel* refSide;
    Pel  refAbove[2*MAX_CU_SIZE+1];
    Pel  refLeft[2*MAX_CU_SIZE+1];

    // Initialise the Main and Left reference array.
    if (intraPredAngle < 0)
    {
      for (k=0;k<blkSize+1;k++)
      {
        refAbove[k+blkSize-1] = pSrc[k-srcStride-1];
      }
      for (k=0;k<blkSize+1;k++)
      {
        refLeft[k+blkSize-1] = pSrc[(k-1)*srcStride-1];
      }
      refMain = (modeVer ? refAbove : refLeft) + (blkSize-1);
      refSide = (modeVer ? refLeft : refAbove) + (blkSize-1);

      // Extend the Main reference to the left.
      Int invAngleSum    = 128;       // rounding for (shift by 8)
      for (k=-1; k>blkSize*intraPredAngle>>5; k--)
      {
        invAngleSum += invAngle;
        refMain[k] = refSide[invAngleSum>>8];
      }
    }
    else
    {
      for (k=0;k<2*blkSize+1;k++)
      {
        refAbove[k] = pSrc[k-srcStride-1];
      }
      for (k=0;k<2*blkSize+1;k++)
      {
        refLeft[k] = pSrc[(k-1)*srcStride-1];
      }
      refMain = modeVer ? refAbove : refLeft;
      refSide = modeVer ? refLeft  : refAbove;
    }

    if (intraPredAngle == 0)
    {
      for (k=0;k<blkSize;k++)
      {
        for (l=0;l<blkSize;l++)
        {
          pDst[k*dstStride+l] = refMain[l+1];
        }
      }

      if ( bFilter )
      {
        for (k=0;k<blkSize;k++)
        {
          pDst[k*dstStride] = Clip ( pDst[k*dstStride] + (( refSide[k+1] - refSide[0] ) >> 1) );
        }
      }
    }
    else
    {
      Int deltaPos=0;
      Int deltaInt;
      Int deltaFract;
      Int refMainIndex;

      for (k=0;k<blkSize;k++)
      {
        deltaPos += intraPredAngle;
        deltaInt   = deltaPos >> 5;
        deltaFract = deltaPos & (32 - 1);

        if (deltaFract)
        {
          // Do linear filtering
          for (l=0;l<blkSize;l++)
          {
            refMainIndex        = l+deltaInt+1;
            pDst[k*dstStride+l] = (Pel) ( ((32-deltaFract)*refMain[refMainIndex]+deltaFract*refMain[refMainIndex+1]+16) >> 5 );
          }
        }
        else
        {
          // Just copy the integer samples
          for (l=0;l<blkSize;l++)
          {
            pDst[k*dstStride+l] = refMain[l+deltaInt+1];
          }
        }
      }
    }

    // Flip the block if this is the horizontal mode
    if (modeHor)
    {
      Pel  tmp;
      for (k=0;k<blkSize-1;k++)
      {
        for (l=k+1;l<blkSize;l++)
        {
          tmp                 = pDst[k*dstStride+l];
          pDst[k*dstStride+l] = pDst[l*dstStride+k];
          pDst[l*dstStride+k] = tmp;
        }
      }
    }
  }
}

Void TComPrediction::predIntraLumaAng(TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft )
{
  Pel *pDst = piPred;
  Int *ptrSrc;

  assert( g_aucConvertToBit[ iWidth ] >= 0 ); //   4x  4
  assert( g_aucConvertToBit[ iWidth ] <= 5 ); // 128x128
  assert( iWidth == iHeight  );

  ptrSrc = pcTComPattern->getPredictorPtr( uiDirMode, g_aucConvertToBit[ iWidth ] + 2, m_piYuvExt );

  // get starting pixel in block
  Int sw = 2 * iWidth + 1;

  // Create the prediction
  if ( uiDirMode == PLANAR_IDX )
  {
    xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight );
  }
  else
  {
    xPredIntraAng( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, true );

    if( (uiDirMode == DC_IDX ) && bAbove && bLeft )
    {
      xDCPredFiltering( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight);
    }
  }
}

// Angular chroma
Void TComPrediction::predIntraChromaAng( TComPattern* pcTComPattern, Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, TComDataCU* pcCU, Bool bAbove, Bool bLeft )
{
  Pel *pDst = piPred;
  Int *ptrSrc = piSrc;

  // get starting pixel in block
  Int sw = 2 * iWidth + 1;

  if ( uiDirMode == PLANAR_IDX )
  {
    xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight );
  }
  else
  {
    // Create the prediction
    xPredIntraAng( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, false );
  }
}

/** Function for checking identical motion.
 * \param TComDataCU* pcCU
 * \param UInt PartAddr
 */
Bool TComPrediction::xCheckIdenticalMotion ( TComDataCU* pcCU, UInt PartAddr )
{
  if( pcCU->getSlice()->isInterB() && !pcCU->getSlice()->getPPS()->getWPBiPred() )
  {
    if( pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr) >= 0 && pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr) >= 0)
    {
      Int RefPOCL0 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_0, pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr))->getPOC();
      Int RefPOCL1 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_1, pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr))->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcCU->getCUMvField(REF_PIC_LIST_0)->getMv(PartAddr) == pcCU->getCUMvField(REF_PIC_LIST_1)->getMv(PartAddr))
      {
        return true;
      }
    }
  }
  return false;
}

#if INTRA_BL
Void TComPrediction::getBaseBlk( TComDataCU* pcCU, TComYuv* pcYuvPred, Int iPartAddr, Int iWidth, Int iHeight )
{
  pcCU->getBaseLumaBlk( iWidth, iHeight, iPartAddr, pcYuvPred->getLumaAddr( iPartAddr ), pcYuvPred->getStride() );
  pcCU->getBaseChromaBlk( iWidth >> 1, iHeight >> 1, iPartAddr, pcYuvPred->getCbAddr( iPartAddr ), pcYuvPred->getCStride(), 0 );
  pcCU->getBaseChromaBlk( iWidth >> 1, iHeight >> 1, iPartAddr, pcYuvPred->getCrAddr( iPartAddr ), pcYuvPred->getCStride(), 1 );
}
#endif

Void TComPrediction::motionCompensation ( TComDataCU* pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList, Int iPartIdx )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  if ( iPartIdx >= 0 )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );
#if SHVC_COMPLEXITY_ASSESSMENT
	// count the number of samples predicted with interpolation
 g_d_num_pel_predicted_with_interpolation +=iHeight*iWidth; 
#endif  //SHVC_COMPLEXITY_ASSESSMENT
    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, iPartIdx, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, iPartIdx );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, iPartIdx );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred, iPartIdx );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred, iPartIdx );
      }
    }
    return;
  }

  for ( iPartIdx = 0; iPartIdx < pcCU->getNumPartInter(); iPartIdx++ )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );
#if SHVC_COMPLEXITY_ASSESSMENT
	// count the number of samples predicted with interpolation
 g_d_num_pel_predicted_with_interpolation +=iHeight*iWidth; 
#endif  //SHVC_COMPLEXITY_ASSESSMENT
    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, iPartIdx, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, iPartIdx );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, iPartIdx );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred, iPartIdx );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred, iPartIdx );
      }
    }
  }
  return;
}

Void TComPrediction::xPredInterUni ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Int iPartIdx, Bool bi )
{
  Int         iRefIdx     = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );           assert (iRefIdx >= 0);
  TComMv      cMv         = pcCU->getCUMvField( eRefPicList )->getMv( uiPartAddr );
  pcCU->clipMv(cMv);
  xPredInterLumaBlk  ( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
  xPredInterChromaBlk( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
#if SHVC_COMPLEXITY_ASSESSMENT
#if REF_IDX_FRAMEWORK
  if(pic_level_calc_done==0 && pcCU->getLayerId())
  {
	  int i,j,k,iWidthCTU,iHeightCTU;
	  int PicWidth=pcCU->getPic()->getPicYuvRec()->getWidth();
	  int PicHeight=pcCU->getPic()->getPicYuvRec()->getHeight();

	  for(j=PicHeight; j>0; j-=64)
	  {
		  iHeightCTU=min(64,j);
		  for(i=PicWidth; i>0; i-=64)
		  {
			  iWidthCTU=min(64,i);	
        if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
        {
			  g_d_num_mults_pic+= iWidthCTU* (iHeightCTU+LUMA_USF_TAPS-1)* LUMA_USF_TAPS; 
			  g_d_num_adds_pic+=iWidthCTU* (iHeightCTU+LUMA_USF_TAPS-1)*(LUMA_USF_TAPS-1);

			  g_d_num_mults_pic+= iWidthCTU* iHeightCTU* LUMA_USF_TAPS; 
			  g_d_num_adds_pic+=iWidthCTU* iHeightCTU*(LUMA_USF_TAPS-1);

			  g_d_num_mults_pic+= 2*iWidthCTU/2* (iHeightCTU/2+CHROMA_USF_TAPS-1)* CHROMA_USF_TAPS; 
			  g_d_num_adds_pic+=2*iWidthCTU/2* (iHeightCTU/2+CHROMA_USF_TAPS-1)*(CHROMA_USF_TAPS-1);

			  g_d_num_mults_pic+=2* iWidthCTU/2* iHeightCTU/2* CHROMA_USF_TAPS; 
			  g_d_num_adds_pic+=2*iWidthCTU/2* iHeightCTU/2*(CHROMA_USF_TAPS-1);
        }
			  Int PosX = PicWidth-i;
			  Int PosY = PicHeight-j;

			  initMCMemoryAccessCalculator( pcCU->getSlice()->getBaseColPic(), 2 ); // for picture-based up-sampling
				if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
				{
				for ( k=0; k<NUM_MEMORY_ARCHITECTURES; k++ )
					{
						if (m_apcMemAccessCalcPic[k])
						{
							m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW), int(iHeightCTU*g_d_ScalingFactorH), int(PosX*g_d_ScalingFactorW), int(PosY*g_d_ScalingFactorH), LUMA_USF_TAPS, LUMA_USF_TAPS, false);
							m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW/2), int(iHeightCTU*g_d_ScalingFactorH/2), int(PosX*g_d_ScalingFactorW/2), int(PosY*g_d_ScalingFactorH/2), CHROMA_USF_TAPS, CHROMA_USF_TAPS, true);
						}
					}
				}
				else  // SNR case
				{
				for ( k=0; k<NUM_MEMORY_ARCHITECTURES; k++ )
					{
						if (m_apcMemAccessCalcPic[k])
						{
							m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW), int(iHeightCTU*g_d_ScalingFactorH), int(PosX*g_d_ScalingFactorW), int(PosY*g_d_ScalingFactorH), 1, 1, false);
							m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW/2), int(iHeightCTU*g_d_ScalingFactorH/2), int(PosX*g_d_ScalingFactorW/2), int(PosY*g_d_ScalingFactorH/2), 1, 1, true);
						}
					}
				}
		  }
	  }
	  pic_level_calc_done=1;
  }
  if (pcCU->getLayerId() && pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->getIsILR() /* detect inter-layer prediction PU*/)
  {  
	  initMCMemoryAccessCalculator( pcCU->getSlice()->getBaseColPic(), 1 ); // for block-base up-sampling
      xCalcUSMemoryAccessLuma(pcCU, uiPartAddr, iWidth, iHeight, LUMA_USF_TAPS, LUMA_USF_TAPS);
      xCalcUSMemoryAccessChroma(pcCU, uiPartAddr, iWidth, iHeight, CHROMA_USF_TAPS, CHROMA_USF_TAPS);
#if ILR_TM_ENH_COMPLEXITY
    if(pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->getIsEnhILR())
    {
      calcEnhILRReconBlock(pcCU, uiPartAddr, iWidth, iHeight);
    }
#endif
  }
  else
  {
#endif
  initMCMemoryAccessCalculator( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx), 0 ); // for MC
  xCalcMCMemoryAccessLuma  ( pcCU, uiPartAddr, eRefPicList, iRefIdx, &cMv, iWidth, iHeight, NTAPS_LUMA, NTAPS_LUMA);
  xCalcMCMemoryAccessChroma( pcCU, uiPartAddr, eRefPicList, iRefIdx, &cMv, iWidth, iHeight, NTAPS_CHROMA, NTAPS_CHROMA);
#if REF_IDX_FRAMEWORK
  }
#endif
#endif //SHVC_COMPLEXITY_ASSESSMENT
#if SHVC_COMPLEXITY_ASSESSMENT && REF_IDX_FRAMEWORK 
  if (pcCU->getLayerId() && pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->getIsILR() /* detect inter-layer prediction PU*/)
  {
if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
{
	// horizontal Luma interpolation
	g_d_num_mults_block+= iWidth* (iHeight+LUMA_USF_TAPS-1)* LUMA_USF_TAPS; 
	g_d_num_adds_block+=iWidth* (iHeight+LUMA_USF_TAPS-1)*(LUMA_USF_TAPS-1);
	 // vertical Luma interpolation
	g_d_num_mults_block+= iWidth* iHeight* LUMA_USF_TAPS; 
	g_d_num_adds_block+=iWidth* iHeight*(LUMA_USF_TAPS-1);
	 // horizontal Chroma interpolation
	g_d_num_mults_block+= 2*iWidth/2* (iHeight/2+CHROMA_USF_TAPS-1)* CHROMA_USF_TAPS; 
	g_d_num_adds_block+=2*iWidth/2* (iHeight/2+CHROMA_USF_TAPS-1)*(CHROMA_USF_TAPS-1);
	 // vertical Chroma interpolation
	g_d_num_mults_block+=2* iWidth/2* iHeight/2* CHROMA_USF_TAPS; 
	g_d_num_adds_block+=2*iWidth/2* iHeight/2*(CHROMA_USF_TAPS-1);
}

  }
#endif //SHVC_COMPLEXITY_ASSESSMENT
}

Void TComPrediction::xPredInterBi ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred, Int iPartIdx )
{
  TComYuv* pcMbYuv;
  Int      iRefIdx[2] = {-1, -1};

  for ( Int iRefList = 0; iRefList < 2; iRefList++ )
  {
    RefPicList eRefPicList = (iRefList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
    iRefIdx[iRefList] = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );

    if ( iRefIdx[iRefList] < 0 )
    {
      continue;
    }

    assert( iRefIdx[iRefList] < pcCU->getSlice()->getNumRefIdx(eRefPicList) );

    pcMbYuv = &m_acYuvPred[iRefList];
    if( pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr ) >= 0 && pcCU->getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiPartAddr ) >= 0 )
    {
      xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, iPartIdx, true );
    }
    else
    {
      if ( ( pcCU->getSlice()->getPPS()->getUseWP()       && pcCU->getSlice()->getSliceType() == P_SLICE ) || 
           ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE ) )
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, iPartIdx, true );
      }
      else
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, iPartIdx );
      }
    }
  }

  if ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE  )
  {
    xWeightedPredictionBi( pcCU, &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }  
  else if ( pcCU->getSlice()->getPPS()->getUseWP() && pcCU->getSlice()->getSliceType() == P_SLICE )
  {
    xWeightedPredictionUni( pcCU, &m_acYuvPred[0], uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, rpcYuvPred, iPartIdx ); 
  }
  else
  {
    xWeightedAverage( pcCU, &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
}

/**
 * \brief Generate motion-compensated luma block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */
Void TComPrediction::xPredInterLumaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int refStride = refPic->getStride();  
  Int refOffset = ( mv->getHor() >> 2 ) + ( mv->getVer() >> 2 ) * refStride;
  Pel *ref      = refPic->getLumaAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Int dstStride = dstPic->getStride();
  Pel *dst      = dstPic->getLumaAddr( partAddr );
  
  Int xFrac = mv->getHor() & 0x3;
  Int yFrac = mv->getVer() & 0x3;

  if ( yFrac == 0 )
  {
    m_if.filterHorLuma( ref, refStride, dst, dstStride, width, height, xFrac,       !bi );
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerLuma( ref, refStride, dst, dstStride, width, height, yFrac, true, !bi );
  }
  else
  {
    Int tmpStride = m_filteredBlockTmp[0].getStride();
    Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

    Int filterSize = NTAPS_LUMA;
    Int halfFilterSize = ( filterSize >> 1 );

    m_if.filterHorLuma(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, xFrac, false     );
    m_if.filterVerLuma(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              yFrac, false, !bi);    
  }
}
#if ILR_TM_ENH
Void TComPrediction::xPredInterLumaBlkByPos(TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int refStride = refPic->getStride();  
  Int refOffset = ( mv->getHor() >> 2 ) + ( mv->getVer() >> 2 ) * refStride;
  Pel *ref      = refPic->getLumaAddr() + iBlkY*refStride + iBlkX + refOffset;
  
  Int dstStride = dstPic->getStride();
  Pel *dst      = dstPic->getLumaAddr( 0 );
  
  Int xFrac = mv->getHor() & 0x3;
  Int yFrac = mv->getVer() & 0x3;

  if ( yFrac == 0 )
  {
    m_if.filterHorLuma( ref, refStride, dst, dstStride, width, height, xFrac,       !bi );
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerLuma( ref, refStride, dst, dstStride, width, height, yFrac, true, !bi );
  }
  else
  {
    Int tmpStride = m_filteredBlockTmp[0].getStride();
    Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

    Int filterSize = NTAPS_LUMA;
    Int halfFilterSize = ( filterSize >> 1 );

    m_if.filterHorLuma(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, xFrac, false     );
    m_if.filterVerLuma(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              yFrac, false, !bi);    
  }
}
#endif
#if ILR_TM_ENH_MC_BILINEAR
Void TComPrediction::xPredInterLumaBlkByPosBilinear(TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int refStride = refPic->getStride();  
  Int refOffset = ( mv->getHor() >> 2 ) + ( mv->getVer() >> 2 ) * refStride;
  Pel *ref      = refPic->getLumaAddr() + iBlkY*refStride + iBlkX + refOffset;
  
  Int dstStride = dstPic->getStride();
  Pel *dst      = dstPic->getLumaAddr( 0 );
  
  Int xFrac = mv->getHor() & 0x3;
  Int yFrac = mv->getVer() & 0x3;

  if ( yFrac == 0 )
  {
    m_if.filterHorBilinear( ref, refStride, dst, dstStride, width, height, (xFrac<<1),       !bi );
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerBilinear( ref, refStride, dst, dstStride, width, height, (yFrac<<1), true, !bi );
  }
  else
  {
    Int tmpStride = m_filteredBlockTmp[0].getStride();
    Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

    Int filterSize = NTAPS_BILINEAR;
    Int halfFilterSize = ( filterSize >> 1 );

    m_if.filterHorBilinear(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, (xFrac<<1), false     );
    m_if.filterVerBilinear(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              (yFrac<<1), false, !bi);    
  }
}
#endif
/**
 * \brief Generate motion-compensated chroma block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */
Void TComPrediction::xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getCStride();
  Int     dstStride  = dstPic->getCStride();
  
  Int     refOffset  = (mv->getHor() >> 3) + (mv->getVer() >> 3) * refStride;
  
  Pel*    refCb     = refPic->getCbAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  Pel*    refCr     = refPic->getCrAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Pel* dstCb = dstPic->getCbAddr( partAddr );
  Pel* dstCr = dstPic->getCrAddr( partAddr );
  
  Int     xFrac  = mv->getHor() & 0x7;
  Int     yFrac  = mv->getVer() & 0x7;
  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;
  
  Int     extStride = m_filteredBlockTmp[0].getStride();
  Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();
  
  Int filterSize = NTAPS_CHROMA;
  
  Int halfFilterSize = (filterSize>>1);
  
  if ( yFrac == 0 )
  {
    m_if.filterHorChroma(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
    m_if.filterHorChroma(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerChroma(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
    m_if.filterVerChroma(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
  }
  else
  {
    m_if.filterHorChroma(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);
    
    m_if.filterHorChroma(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
  }
}
#if ILR_TM_ENH
Void TComPrediction::xPredInterChromaBlkByPos(TComPicYuv *refPic, Int iBlkX, Int iBlkY, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getCStride();
  Int     dstStride  = dstPic->getCStride();
  
  Int     refOffset  = (mv->getHor() >> 3) + (mv->getVer() >> 3) * refStride;
  
  Pel*    refCb     = refPic->getCbAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
  Pel*    refCr     = refPic->getCrAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
  
  Pel* dstCb = dstPic->getCbAddr( 0 );
  Pel* dstCr = dstPic->getCrAddr( 0 );
  
  Int     xFrac  = mv->getHor() & 0x7;
  Int     yFrac  = mv->getVer() & 0x7;
  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;
  
  Int     extStride = m_filteredBlockTmp[0].getStride();
  Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();
  
  Int filterSize = NTAPS_CHROMA;
  
  Int halfFilterSize = (filterSize>>1);
  
  if ( yFrac == 0 )
  {
    m_if.filterHorChroma(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
    m_if.filterHorChroma(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerChroma(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
    m_if.filterVerChroma(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
  }
  else
  {
    m_if.filterHorChroma(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);
    
    m_if.filterHorChroma(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
  }
}
#endif
#if ILR_TM_ENH_MC_BILINEAR
Void TComPrediction::xPredInterChromaBlkByPosBilinear(TComPicYuv *refPic, Int iBlkX, Int iBlkY, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getCStride();
  Int     dstStride  = dstPic->getCStride();
  
  Int     refOffset  = (mv->getHor() >> 3) + (mv->getVer() >> 3) * refStride;
  
  Pel*    refCb     = refPic->getCbAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
  Pel*    refCr     = refPic->getCrAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
  
  Pel* dstCb = dstPic->getCbAddr( 0 );
  Pel* dstCr = dstPic->getCrAddr( 0 );
  
  Int     xFrac  = mv->getHor() & 0x7;
  Int     yFrac  = mv->getVer() & 0x7;
  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;
  
  Int     extStride = m_filteredBlockTmp[0].getStride();
  Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();
  
  Int filterSize = NTAPS_BILINEAR;
  
  Int halfFilterSize = (filterSize>>1);
  
  if ( yFrac == 0 )
  {
    m_if.filterHorBilinear(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
    m_if.filterHorBilinear(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerBilinear(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
    m_if.filterVerBilinear(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
  }
  else
  {
    m_if.filterHorBilinear(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerBilinear(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);
    
    m_if.filterHorBilinear(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerBilinear(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
  }
}
#endif
Void TComPrediction::xWeightedAverage( TComDataCU* pcCU, TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartIdx, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst )
{
  if( iRefIdx0 >= 0 && iRefIdx1 >= 0 )
  {
    rpcYuvDst->addAvg( pcYuvSrc0, pcYuvSrc1, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 >= 0 && iRefIdx1 <  0 )
  {
    pcYuvSrc0->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 <  0 && iRefIdx1 >= 0 )
  {
    pcYuvSrc1->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
}

// AMVP
Void TComPrediction::getMvPredAMVP( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv& rcMvPred )
{
  AMVPInfo* pcAMVPInfo = pcCU->getCUMvField(eRefPicList)->getAMVPInfo();
#if !SPS_AMVP_CLEANUP
  if( pcCU->getAMVPMode(uiPartAddr) == AM_NONE || (pcAMVPInfo->iN <= 1 && pcCU->getAMVPMode(uiPartAddr) == AM_EXPL) )
#else
  if( pcAMVPInfo->iN <= 1 )
#endif
  {
    rcMvPred = pcAMVPInfo->m_acMvCand[0];

    pcCU->setMVPIdxSubParts( 0, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    pcCU->setMVPNumSubParts( pcAMVPInfo->iN, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    return;
  }

  assert(pcCU->getMVPIdx(eRefPicList,uiPartAddr) >= 0);
  rcMvPred = pcAMVPInfo->m_acMvCand[pcCU->getMVPIdx(eRefPicList,uiPartAddr)];
  return;
}

/** Function for deriving planar intra prediction.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 *
 * This function derives the prediction samples for planar mode (intra coding).
 */
Void TComPrediction::xPredIntraPlanar( Int* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height )
{
  assert(width == height);

  Int k, l, bottomLeft, topRight;
  Int horPred;
  Int leftColumn[MAX_CU_SIZE], topRow[MAX_CU_SIZE], bottomRow[MAX_CU_SIZE], rightColumn[MAX_CU_SIZE];
  UInt blkSize = width;
  UInt offset2D = width;
  UInt shift1D = g_aucConvertToBit[ width ] + 2;
  UInt shift2D = shift1D + 1;

  // Get left and above reference column and row
  for(k=0;k<blkSize+1;k++)
  {
    topRow[k] = pSrc[k-srcStride];
    leftColumn[k] = pSrc[k*srcStride-1];
  }

  // Prepare intermediate variables used in interpolation
  bottomLeft = leftColumn[blkSize];
  topRight   = topRow[blkSize];
  for (k=0;k<blkSize;k++)
  {
    bottomRow[k]   = bottomLeft - topRow[k];
    rightColumn[k] = topRight   - leftColumn[k];
    topRow[k]      <<= shift1D;
    leftColumn[k]  <<= shift1D;
  }

  // Generate prediction signal
  for (k=0;k<blkSize;k++)
  {
    horPred = leftColumn[k] + offset2D;
    for (l=0;l<blkSize;l++)
    {
      horPred += rightColumn[k];
      topRow[l] += bottomRow[l];
      rpDst[k*dstStride+l] = ( (horPred + topRow[l]) >> shift2D );
    }
  }
}

#if !REMOVE_LMCHROMA
/** Function for deriving chroma LM intra prediction.
 * \param pcPattern pointer to neighbouring pixel access pattern
 * \param piSrc pointer to reconstructed chroma sample array
 * \param pPred pointer for the prediction sample array
 * \param uiPredStride the stride of the prediction sample array
 * \param uiCWidth the width of the chroma block
 * \param uiCHeight the height of the chroma block
 * \param uiChromaId boolean indication of chroma component
 *
 * This function derives the prediction samples for chroma LM mode (chroma intra coding)
 */
Void TComPrediction::predLMIntraChroma( TComPattern* pcPattern, Int* piSrc, Pel* pPred, UInt uiPredStride, UInt uiCWidth, UInt uiCHeight, UInt uiChromaId )
{
  UInt uiWidth  = 2 * uiCWidth;

  xGetLLSPrediction( pcPattern, piSrc+uiWidth+2, uiWidth+1, pPred, uiPredStride, uiCWidth, uiCHeight, 1 );  
}

/** Function for deriving downsampled luma sample of current chroma block and its above, left causal pixel
 * \param pcPattern pointer to neighbouring pixel access pattern
 * \param uiCWidth the width of the chroma block
 * \param uiCHeight the height of the chroma block
 *
 * This function derives downsampled luma sample of current chroma block and its above, left causal pixel
 */
Void TComPrediction::getLumaRecPixels( TComPattern* pcPattern, UInt uiCWidth, UInt uiCHeight )
{
  UInt uiWidth  = 2 * uiCWidth;
  UInt uiHeight = 2 * uiCHeight;  

  Pel* pRecSrc = pcPattern->getROIY();
  Pel* pDst0 = m_pLumaRecBuffer + m_iLumaRecStride + 1;

  Int iRecSrcStride = pcPattern->getPatternLStride();
  Int iRecSrcStride2 = iRecSrcStride << 1;
  Int iDstStride = m_iLumaRecStride;
  Int iSrcStride = ( max( uiWidth, uiHeight ) << 1 ) + 1;

  Int* ptrSrc = pcPattern->getAdiOrgBuf( uiWidth, uiHeight, m_piYuvExt );

  // initial pointers
  Pel* pDst = pDst0 - 1 - iDstStride;  
  Int* piSrc = ptrSrc;

  // top left corner downsampled from ADI buffer
  // don't need this point

  // top row downsampled from ADI buffer
  pDst++;     
  piSrc ++;
  for (Int i = 0; i < uiCWidth; i++)
  {
    pDst[i] = ((piSrc[2*i] * 2 ) + piSrc[2*i - 1] + piSrc[2*i + 1] + 2) >> 2;
  }

  // left column downsampled from ADI buffer
  pDst = pDst0 - 1; 
  piSrc = ptrSrc + iSrcStride;
  for (Int j = 0; j < uiCHeight; j++)
  {
    pDst[0] = ( piSrc[0] + piSrc[iSrcStride] ) >> 1;
    piSrc += iSrcStride << 1; 
    pDst += iDstStride;    
  }

  // inner part from reconstructed picture buffer
  for( Int j = 0; j < uiCHeight; j++ )
  {
    for (Int i = 0; i < uiCWidth; i++)
    {
      pDst0[i] = (pRecSrc[2*i] + pRecSrc[2*i + iRecSrcStride]) >> 1;
    }

    pDst0 += iDstStride;
    pRecSrc += iRecSrcStride2;
  }
}

/** Function for deriving the positon of first non-zero binary bit of a value
 * \param x input value
 *
 * This function derives the positon of first non-zero binary bit of a value
 */
Int GetFloorLog2( UInt x )
{
  int bits = -1;
  while( x > 0 )
  {
    bits ++;
    x >>= 1;
  }
  return bits;
}

/** Function for deriving LM intra prediction.
 * \param pcPattern pointer to neighbouring pixel access pattern
 * \param pSrc0 pointer to reconstructed chroma sample array
 * \param iSrcStride the stride of reconstructed chroma sample array
 * \param pDst0 reference to pointer for the prediction sample array
 * \param iDstStride the stride of the prediction sample array
 * \param uiWidth the width of the chroma block
 * \param uiHeight the height of the chroma block
 * \param uiExt0 line number of neiggboirng pixels for calculating LM model parameter, default value is 1
 *
 * This function derives the prediction samples for chroma LM mode (chroma intra coding)
 */
Void TComPrediction::xGetLLSPrediction( TComPattern* pcPattern, Int* pSrc0, Int iSrcStride, Pel* pDst0, Int iDstStride, UInt uiWidth, UInt uiHeight, UInt uiExt0 )
{

  Pel  *pDst, *pLuma;
  Int  *pSrc;

  Int  iLumaStride = m_iLumaRecStride;
  Pel* pLuma0 = m_pLumaRecBuffer + uiExt0 * iLumaStride + uiExt0;

  Int i, j, iCountShift = 0;
  UInt uiInternalBitDepth = g_uiBitDepth + g_uiBitIncrement;

  UInt uiExt = uiExt0;

  // LLS parameters estimation -->

  Int x = 0, y = 0, xx = 0, xy = 0;

  pSrc  = pSrc0  - iSrcStride;
  pLuma = pLuma0 - iLumaStride;

  for( j = 0; j < uiWidth; j++ )
  {
    x += pLuma[j];
    y += pSrc[j];
    xx += pLuma[j] * pLuma[j];
    xy += pLuma[j] * pSrc[j];
  }

  pSrc  = pSrc0 - uiExt;
  pLuma = pLuma0 - uiExt;

  for( i = 0; i < uiHeight; i++ )
  {
    x += pLuma[0];
    y += pSrc[0];
    xx += pLuma[0] * pLuma[0];
    xy += pLuma[0] * pSrc[0];

    pSrc  += iSrcStride;
    pLuma += iLumaStride;
  }
  iCountShift = g_aucConvertToBit[ uiWidth ] + 3;

  Int iTempShift = uiInternalBitDepth + iCountShift - 15;

  if(iTempShift > 0)
  {
    x  = ( x +  ( 1 << ( iTempShift - 1 ) ) ) >> iTempShift;
    y  = ( y +  ( 1 << ( iTempShift - 1 ) ) ) >> iTempShift;
    xx = ( xx + ( 1 << ( iTempShift - 1 ) ) ) >> iTempShift;
    xy = ( xy + ( 1 << ( iTempShift - 1 ) ) ) >> iTempShift;
    iCountShift -= iTempShift;
  }

  Int avgLuma =  x   >> iCountShift;
  Int avgSrc =  y  >> iCountShift;
  Int RErrLuma = x & ( ( 1 << iCountShift ) - 1 );
  Int RErrSrc =  y & ( ( 1 << iCountShift ) - 1 );

  Int a, b, iShift = 13;

  Int iB = 7;
  iShift -= iB;

  if( iCountShift == 0 )
  {
    a = 0;
    b = 1 << (uiInternalBitDepth - 1);
    iShift = 0;
  }
  else
  {
    Int a1 = xy - ( avgLuma*avgSrc << iCountShift ) - avgLuma*RErrSrc - avgSrc*RErrLuma;
    Int a2 = xx - ( avgLuma*avgLuma << iCountShift ) - 2*avgLuma*RErrLuma;

    const Int iShiftA1 = uiInternalBitDepth - 2;
    const Int iShiftA2 = 5;
    const Int iAccuracyShift = uiInternalBitDepth + 4;

    Int iScaleShiftA2 = 0;
    Int iScaleShiftA1 = 0;
    Int a1s = a1;
    Int a2s = a2;

    iScaleShiftA1 = a1 == 0 ? 0 : GetFloorLog2( abs( a1 ) ) - iShiftA1;
    iScaleShiftA2 = a2 == 0 ? 0 : GetFloorLog2( abs( a2 ) ) - iShiftA2;

    if( iScaleShiftA1 < 0 )
    {
      iScaleShiftA1 = 0;
    }
    
    if( iScaleShiftA2 < 0 )
    {
      iScaleShiftA2 = 0;
    }

    Int iScaleShiftA = iScaleShiftA2 + iAccuracyShift - iShift - iScaleShiftA1;

    a2s = a2 >> iScaleShiftA2;

    a1s = a1 >> iScaleShiftA1;

    if (a2s >= 32)
    {
      UInt a2t = m_uiaShift[ a2s - 32 ] ;
      a2t = Clip( a2t );
      a = a1s * a2t;
    }
    else
    {
      a = 0;
    }
    
    if( iScaleShiftA < 0 )
    {
      a = a << -iScaleShiftA;
    }
    else
    {
      a = a >> iScaleShiftA;
    }
    a = Clip3(-( 1 << (15-iB) ), ( 1 << (15-iB )) - 1, a);
    a = a << iB;
   
    Short n = 0;
    if (a != 0)
    {
      n = GetFloorLog2(abs( a ) + ( (a < 0 ? -1 : 1) - 1)/2 ) - 5;
    }
    
    iShift =(iShift+iB)-n;
    a = a>>n;

    b =  avgSrc - ( (  a * avgLuma ) >> iShift );
  }   

  // <-- end of LLS parameters estimation

  // get prediction -->
  uiExt = uiExt0;
  pLuma = pLuma0;
  pDst = pDst0;

  for( i = 0; i < uiHeight; i++ )
  {
    for( j = 0; j < uiWidth; j++ )
    {
      pDst[j] = Clip( ( ( a * pLuma[j] ) >> iShift ) + b );
    }
    
    pDst  += iDstStride;
    pLuma += iLumaStride;
  }
  // <-- end of get prediction

}
#endif

/** Function for filtering intra DC predictor.
 * \param pSrc pointer to reconstructed sample array
 * \param iSrcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param iDstStride the stride of the prediction sample array
 * \param iWidth the width of the block
 * \param iHeight the height of the block
 *
 * This function performs filtering left and top edges of the prediction samples for DC mode (intra coding).
 */
Void TComPrediction::xDCPredFiltering( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight )
{
  Pel* pDst = rpDst;
  Int x, y, iDstStride2, iSrcStride2;

  // boundary pixels processing
  pDst[0] = (Pel)((pSrc[-iSrcStride] + pSrc[-1] + 2 * pDst[0] + 2) >> 2);

  for ( x = 1; x < iWidth; x++ )
  {
    pDst[x] = (Pel)((pSrc[x - iSrcStride] +  3 * pDst[x] + 2) >> 2);
  }

  for ( y = 1, iDstStride2 = iDstStride, iSrcStride2 = iSrcStride-1; y < iHeight; y++, iDstStride2+=iDstStride, iSrcStride2+=iSrcStride )
  {
    pDst[iDstStride2] = (Pel)((pSrc[iSrcStride2] + 3 * pDst[iDstStride2] + 2) >> 2);
  }

  return;
}

#if SVC_UPSAMPLING
Void TComPrediction::upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic)
{
  m_cUsf.upsampleBasePic( pcUsPic, pcBasePic, pcTempPic);
}
#endif

#if SHVC_COMPLEXITY_ASSESSMENT 
/*
 * \brief Initialize parameters of an MC memory access computation module.
 *
 * \param pcRefPic
 *
 * \return void
 */

Void TComPrediction::initMCMemoryAccessCalculator( TComPic* pcRefPic, int design_flag )
{
#if ILR_TM_ENH_COMPLEXITY
	if(design_flag == 3)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalcPic_enhILR[i])
			{
				m_apcMemAccessCalcPic_enhILR[i]->setLumaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalcPic_enhILR[i]->setChromaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalcPic_enhILR[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalcPic_enhILR[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
  else
#endif
	if(design_flag == 2)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalcPic[i])
			{
				m_apcMemAccessCalcPic[i]->setLumaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalcPic[i]->setChromaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalcPic[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalcPic[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
	else if(design_flag == 1)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalcBlock[i])
			{
				m_apcMemAccessCalcBlock[i]->setLumaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalcBlock[i]->setChromaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalcBlock[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalcBlock[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
	else
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalc[i])
			{
				m_apcMemAccessCalc[i]->setLumaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalc[i]->setChromaBitDepth( g_uiBitDepth+g_uiBitIncrement );
				m_apcMemAccessCalc[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalc[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
}
Void TComPrediction::xCalcUSMemoryAccessLuma( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for luma. */
  Int iLumaPUPosX = pcCU->getSlice()->getFullPelBaseRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr );
  Int iLumaPUPosY = pcCU->getSlice()->getFullPelBaseRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr );

  Int iMCTapH = iFilTapH;
  Int iMCTapV = iFilTapV;

  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcBlock[i])
    {
		if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
		{
			m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW), int(iHeight*g_d_ScalingFactorH), int(iLumaPUPosX*g_d_ScalingFactorW), int(iLumaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, false);
		}
		else  // SNR case
		{
			m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW), int(iHeight*g_d_ScalingFactorH), int(iLumaPUPosX*g_d_ScalingFactorW), int(iLumaPUPosY*g_d_ScalingFactorH), 1, 1, false);
		}
    }
  }
}
Void TComPrediction::xCalcUSMemoryAccessChroma( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for chroma. */
  Int iChromaPUPosX = pcCU->getSlice()->getFullPelBaseRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  Int iChromaPUPosY = pcCU->getSlice()->getFullPelBaseRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  
  Int iMCTapH = iFilTapH;
  Int iMCTapV = iFilTapV;

  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcBlock[i])
    {
		if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
		{
		      m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, true);
		}
		else // SNR case
		{
		      //m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, true);
		      m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), 1, 1, true);// jdong bug fix
		}
    }
  }
}
/*
 * \brief Compute the amount of MC memory access for a luma block.
 *
 * \param pcCU
 * \param uiPartAddr
 * \param eRefPicList
 * \param iRefIdx
 * \param pcMv
 * \param iWidth
 * \param iHeight
 * \param iFilTapH
 * \param iFilTapV
 *
 * \return void
 */
Void TComPrediction::xCalcMCMemoryAccessLuma( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for luma. */
  Int iLumaPUPosX = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int iLumaPUPosY = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int iLumaPosX = pcMv->getHor() >> 2;
  Int iLumaPosY = pcMv->getVer() >> 2;
  int ixFrac = (pcMv->getHor() & 0x3);
  int iyFrac = (pcMv->getVer() & 0x3);
  if (ixFrac  == 1 || ixFrac  == 3) iFilTapH=iFilTapH-1;
  if (iyFrac  == 1 || iyFrac  == 3) iFilTapV=iFilTapV-1;

  Int iMCTapH = ((pcMv->getHor() & 0x3) != 0) ? iFilTapH : 0;
  Int iMCTapV = ((pcMv->getVer() & 0x3) != 0) ? iFilTapV : 0;


  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalc[i])
    {
      m_apcMemAccessCalc[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
    }
  }
}


/*
 * \brief Compute the amount of MC memory access for a chroma block.
 *
 * \param pcCU
 * \param uiPartAddr
 * \param eRefPicList
 * \param iRefIdx
 * \param pcMv
 * \param iWidth
 * \param iHeight
 * \param iFilTapH
 * \param iFilTapV
 *
 * \return void
 */
Void TComPrediction::xCalcMCMemoryAccessChroma( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for chroma. */
  Int iChromaPUPosX = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  Int iChromaPUPosY = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  Int iChromaPosX = pcMv->getHor() >> 3;
  Int iChromaPosY = pcMv->getVer() >> 3;
  Int iMCTapH = ((pcMv->getHor() & 0x7) != 0) ? iFilTapH : 0;
  Int iMCTapV = ((pcMv->getVer() & 0x7) != 0) ? iFilTapV : 0;

  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalc[i])
    {
      m_apcMemAccessCalc[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
    }
  }
}

/*
 * \brief Initialize parameters of an MC memory access computation module.
 *
 * \param pcRefPic
 *
 * \return void
 */

Void TComPrediction::initMCMemoryAccessCalculator( const MemCmpParam& cLumaParam, const MemCmpParam& cChromaParam, int design_flag)
{
#if ILR_TM_ENH_COMPLEXITY
	if(design_flag == 3)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalcPic_enhILR[i] = pMAC;
			}
			m_apcMemAccessCalcPic_enhILR[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalcPic_enhILR[i]->attachCache( pcCache );
			}
		}
	}
  else 
#endif
	if(design_flag == 2)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalcPic[i] = pMAC;
			}
			m_apcMemAccessCalcPic[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalcPic[i]->attachCache( pcCache );
			}
		}
	}
	else if(design_flag == 1)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalcBlock[i] = pMAC;
			}
			m_apcMemAccessCalcBlock[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalcBlock[i]->attachCache( pcCache );
			}
		}
	}
	else
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalc[i] = pMAC;
			}
			m_apcMemAccessCalc[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalc[i]->attachCache( pcCache );
			}
		}
	}
}

/*
 * \brief Update values of an MC memory access computation module.
 *
 * \return void
 */
Void TComPrediction::updatePic( int design_flag )
{
#if ILR_TM_ENH_COMPLEXITY
	if (design_flag == 3)
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalcPic_enhILR[i] ) m_apcMemAccessCalcPic_enhILR[i]->update();
		}
  else 
#endif
	if (design_flag == 2)
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalcPic[i] ) m_apcMemAccessCalcPic[i]->update();
		}
	else if (design_flag == 1)
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalcBlock[i] ) m_apcMemAccessCalcBlock[i]->update();
		}
	else
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalc[i] ) m_apcMemAccessCalc[i]->update();
		}
}
/*
 * \brief Get the value of total MC memory access measure for a specified memory architecture in byte.
 *
 * \param iIndex Memory archtecture index
 *
 * \return The total MC memory access measure for the specified memory architecture. 
 */

UInt64 TComPrediction::getTotalMCMemoryAccessBytes( Int iIndex, int design_flag  )
{
  assert ( 0 <= iIndex && iIndex < NUM_MEMORY_ARCHITECTURES );
#if ILR_TM_ENH_COMPLEXITY
  if (design_flag == 3)
	  return (m_apcMemAccessCalcPic_enhILR[iIndex]) ? m_apcMemAccessCalcPic_enhILR[iIndex]->getTotalAccessBytes() : 0;
  else 
#endif
  if (design_flag == 2)
	return (m_apcMemAccessCalcPic[iIndex]) ? m_apcMemAccessCalcPic[iIndex]->getTotalAccessBytes() : 0;
  else if (design_flag == 1)
	return (m_apcMemAccessCalcBlock[iIndex]) ? m_apcMemAccessCalcBlock[iIndex]->getTotalAccessBytes() : 0;
  else
	return (m_apcMemAccessCalc[iIndex]) ? m_apcMemAccessCalc[iIndex]->getTotalAccessBytes() : 0;
}

#endif //SHVC_COMPLEXITY_ASSESSMENT

#if ILR_TM_ENH
#if 0
//debug;
static Void outputBlkFromPic(TComPicYuv *pcPicYuvUpBL, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight)
{
  Int iStrideSrc = pcPicYuvUpBL->getStride();
  Pel *pSrc = pcPicYuvUpBL->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
      printf("%8d ", pSrc[i]);
    printf("\n");
    pSrc += iStrideSrc;
  }
}

static Void outputBlk(TComYuv *pPred, UInt uiAbsPartIdx, Int iBlkWidth, Int iBlkHeight)
{
  Int iStrideSrc = pPred->getStride();
  Pel *pSrc = pPred->getLumaAddr(uiAbsPartIdx);

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
      printf("%8d ", pSrc[i]);
    printf("\n");
    pSrc += iStrideSrc;
  }
}
#endif

static Void copyBlk(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight)
{
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));

  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iBlkX >>= 1;
  iBlkY >>= 1;
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  
  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
  pSrc = pcPicYuvSrc->getCrAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
}

Void TComPrediction::calcEnhBlk(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight) //, UInt uiAbsPartIdx)
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
#if !ILR_TM_ENH_MCD
  Int iStrideMCELx = m_pEnhPred[1]->getStride();
#endif
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
#if !ILR_TM_ENH_MCD
  Pel *pMCELx = m_pEnhPred[1]->getLumaAddr(uiAbsPartIdx);
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i+=2)
    {
#if ILR_TM_ENH_MCD
      pDst[i] = Clip(pSrc[i]+pMCUpBLx[i] - DIFFPIC_OFFSET);
      pDst[i+1] = Clip(pSrc[i+1]+pMCUpBLx[i+1] - DIFFPIC_OFFSET);
#else
      pDst[i] = Clip(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
      pDst[i+1] = Clip(pSrc[i+1]+pMCELx[i+1]-pMCUpBLx[i+1]);
#endif
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if !ILR_TM_ENH_MCD
    pMCELx += iStrideMCELx;
#endif
    pMCUpBLx += iStrideMCUpBLx;
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
#if ILR_TM_ENH_CHROMA_ENABLE
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
#endif
#if !ILR_TM_ENH_MCD
  iStrideMCELx = m_pEnhPred[1]->getCStride();
#endif
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
#if ILR_TM_ENH_CHROMA_ENABLE
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
#endif
#if !ILR_TM_ENH_MCD
  pMCELx = m_pEnhPred[1]->getCbAddr(uiAbsPartIdx);
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
#if ILR_TM_ENH_CHROMA_ENABLE
    for(Int i=0; i<iBlkWidth; i++)
    {
#if ILR_TM_ENH_MCD
      pDst[i] = Clip(pSrc[i]+ pMCUpBLx[i] - DIFFPIC_OFFSET);
#else
      pDst[i] = Clip(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
#endif
    }
#else
    memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if !ILR_TM_ENH_MCD
    pMCELx += iStrideMCELx;
#endif
#if ILR_TM_ENH_CHROMA_ENABLE
    pMCUpBLx += iStrideMCUpBLx;
#endif
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
#if ILR_TM_ENH_CHROMA_ENABLE
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
#endif
#if !ILR_TM_ENH_MCD
  pMCELx = m_pEnhPred[1]->getCrAddr(uiAbsPartIdx);
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
#if ILR_TM_ENH_CHROMA_ENABLE
    for(Int i=0; i<iBlkWidth; i++)
    {
#if ILR_TM_ENH_MCD
      pDst[i] = Clip(pSrc[i]+ pMCUpBLx[i] - DIFFPIC_OFFSET);
#else
      pDst[i] = Clip(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
#endif
    }
#else 
    memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if !ILR_TM_ENH_MCD
    pMCELx += iStrideMCELx;
#endif
#if ILR_TM_ENH_CHROMA_ENABLE
    pMCUpBLx += iStrideMCUpBLx;
#endif
  }
}

//compressed MV;
Void TComPrediction::getEnhPUBlock(TComPic *pcPicSrc, UInt uiCUAddr, UInt uiAbsPartIdx, TComPic *pcPicDst, TComMv *pcBlkSize, Int *piSR) 
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();
  TComDataCU *pcILRCU = pcPicDst->getPicSym()->getCU(uiCUAddr);
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  TComPicYuv *pcPicYuvDst = pcPicDst->getPicYuvRec();
  Int iRefIdx0 = pcILRCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiAbsPartIdx);
  Int iRefIdx1 = pcILRCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiAbsPartIdx);
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  Int iBlkX = pcILRCU->getCUPelX()+g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ];
  Int iBlkY = pcILRCU->getCUPelY()+g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ];
  
  if(iRefIdx0 >=0 || iRefIdx1 >=0)
  {
    Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
    if(bi)
    {      
      TComSlice *pcSliceBL = pcSlice->getBaseColPic()->getSlice(pcSlice->getBaseColPic()->getCurrSliceIdx());
      Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
      Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcILRCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiAbsPartIdx) == pcILRCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiAbsPartIdx))
      {
        bi = false;
      }
    }
    if(!bi)
    {
      //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
      RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
      Int        iRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;
      TComMv     cMv     = pcILRCU->getCUMvField( list )->getMv( uiAbsPartIdx );
      TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iRefIdx )->getFullPelBaseRec();
#if !ILR_TM_ENH_MCD
      TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iRefIdx )->getPicYuvRec();
#endif
      pcSlice->clipMv(cMv, iBlkX, iBlkY);
#if ILR_TM_ENH_MC_BILINEAR
      //Up(BLx); 
      //xPredInterLumaBlk  ( pcILRCU, pcUpBLx, uiAbsPartIdx, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
      //xPredInterChromaBlk( pcILRCU, pcUpBLx, uiAbsPartIdx, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#endif
#if !ILR_TM_ENH_MCD
      //ELx
      xPredInterLumaBlkByPosBilinear  ( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#endif
#endif
#else
      //Up(BLx); 
      //xPredInterLumaBlk  ( pcILRCU, pcUpBLx, uiAbsPartIdx, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
      //xPredInterChromaBlk( pcILRCU, pcUpBLx, uiAbsPartIdx, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
      xPredInterLumaBlkByPos  ( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#endif
#if !ILR_TM_ENH_MCD
      //ELx
      xPredInterLumaBlkByPos  ( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#endif
#endif
#endif
    }
    else
    {
      //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
      TComMv     cMv     = pcILRCU->getCUMvField( REF_PIC_LIST_0 )->getMv( uiAbsPartIdx );
      TComMv     cMv1     = pcILRCU->getCUMvField( REF_PIC_LIST_1 )->getMv( uiAbsPartIdx );
      TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
#if !ILR_TM_ENH_MCD
      TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
      TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
#endif
      pcSlice->clipMv(cMv, iBlkX, iBlkY);
      pcSlice->clipMv(cMv1, iBlkX, iBlkY);
#if ILR_TM_ENH_MC_BILINEAR
      //Up(BLx); 
      TComYuv* pcMbYuv = m_acYuvPred;

      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      //m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, uiAbsPartIdx, iBlkWidth, iBlkHeight );
      m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[0]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#if !ILR_TM_ENH_MCD
      //ELx
      pcMbYuv = m_acYuvPred;
      xPredInterLumaBlkByPosBilinear  ( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      xPredInterLumaBlkByPosBilinear  ( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE      
      xPredInterChromaBlkByPosBilinear( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      //m_pEnhPred[1]->addAvg( m_acYuvPred, m_acYuvPred+1, uiAbsPartIdx, iBlkWidth, iBlkHeight );
      m_pEnhPred[1]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[1]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#endif
#else
      //Up(BLx); 
      TComYuv* pcMbYuv = m_acYuvPred;

      xPredInterLumaBlkByPos  ( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      xPredInterLumaBlkByPos  ( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE      
      xPredInterChromaBlkByPos( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      //m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, uiAbsPartIdx, iBlkWidth, iBlkHeight );
      m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[0]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#if !ILR_TM_ENH_MCD
      //ELx
      pcMbYuv = m_acYuvPred;
      xPredInterLumaBlkByPos  ( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      xPredInterLumaBlkByPos  ( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE      
      xPredInterChromaBlkByPos( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      //m_pEnhPred[1]->addAvg( m_acYuvPred, m_acYuvPred+1, uiAbsPartIdx, iBlkWidth, iBlkHeight );
      m_pEnhPred[1]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[1]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#endif
#endif
    }
    //calc;
    //calcEnhBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight, uiAbsPartIdx);
    calcEnhBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
  }
  else
  {
    //pcPicYuvUpBL -> pcPicYuvDst
    copyBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
  }

}
Void TComPrediction::getEnhBlock(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize, Int *piSR)      
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComDataCU *pcBLCU;
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  TComPicYuv *pcPicYuvDst = pcPicDst->getPicYuvRec();
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  Int iRefIdx0, iRefIdx1;

  pcPicBL->deriveUnitIdxBase(iBlkX /*+ (iBlkWidth>>1)*/, iBlkY /*+ (iBlkHeight>>1)*/, *piSR, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcPicBL->getCU(uiBaseCUAddr);
  iRefIdx0 = pcBLCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiBaseAbsPartIdx);
  iRefIdx1 = pcBLCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiBaseAbsPartIdx);  
  
  if(iRefIdx0 >=0 || iRefIdx1 >=0)
  {
    //TComDataCU *pcILRCU = pcPicDst->getPicSym()->getCU(0);
    Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
    if(bi)
    {
      TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
      Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
      Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcBLCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiBaseAbsPartIdx) == pcBLCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiBaseAbsPartIdx))
      {
        bi = false;
      }
    }
    if(!bi)
    {
      //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
      RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
      Int        iRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;
      TComMv     cMv     = pcBLCU->getCUMvField( list )->getMv( uiBaseAbsPartIdx );
      TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iRefIdx )->getFullPelBaseRec();
#if !ILR_TM_ENH_MCD
      TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iRefIdx )->getPicYuvRec();
#endif
      cMv.scale(piSR);
      pcSlice->clipMv(cMv, iBlkX, iBlkY);
#if ILR_TM_ENH_MC_BILINEAR
      //Up(BLx); 
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#endif
#if !ILR_TM_ENH_MCD
      //ELx
      xPredInterLumaBlkByPosBilinear  ( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#endif
#endif
#else
      //Up(BLx); 
      xPredInterLumaBlkByPos  ( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#endif
#if !ILR_TM_ENH_MCD
      //ELx
      xPredInterLumaBlkByPos  ( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcELxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[1], bi );
#endif
#endif
#endif
    }
    else
    {
      //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
      TComMv     cMv     = pcBLCU->getCUMvField( REF_PIC_LIST_0 )->getMv( uiBaseAbsPartIdx );
      TComMv cMv1     = pcBLCU->getCUMvField( REF_PIC_LIST_1 )->getMv( uiBaseAbsPartIdx );
      TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
#if !ILR_TM_ENH_MCD
      TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
      TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
#endif
      cMv.scale(piSR);
      cMv1.scale(piSR);

      pcSlice->clipMv(cMv, iBlkX, iBlkY);
      pcSlice->clipMv(cMv1, iBlkX, iBlkY);
#if ILR_TM_ENH_MC_BILINEAR
      //Up(BLx); 
      TComYuv* pcMbYuv = m_acYuvPred;
      //TComPicYuv *pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[0]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#if !ILR_TM_ENH_MCD      
      //ELx
      pcMbYuv = m_acYuvPred;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
      xPredInterLumaBlkByPosBilinear  ( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
      xPredInterLumaBlkByPosBilinear  ( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      m_pEnhPred[1]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[1]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#endif
#else
      //Up(BLx); 
      TComYuv* pcMbYuv = m_acYuvPred;
      //TComPicYuv *pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      xPredInterLumaBlkByPos  ( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
      xPredInterLumaBlkByPos  ( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[0]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#if !ILR_TM_ENH_MCD      
      //ELx
      pcMbYuv = m_acYuvPred;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
      xPredInterLumaBlkByPos  ( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcELxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
      xPredInterLumaBlkByPos  ( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE      
      xPredInterChromaBlkByPos( pcELxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      m_pEnhPred[1]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[1]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#endif
#endif
    }
    //calc;
    calcEnhBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
  }
  else
  {
    //pcPicYuvUpBL -> pcPicYuvDst
    copyBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
  }
}

Void TComPrediction::enhPicWithTM(TComPic *pcPicSrc, TComPic *pcPicDst, Int *piSR
  )
{
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  UInt uiPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  UInt uiPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();
  //set SPS;
  pcPicDst->getSlice(0)->setSPS(pcSlice->getSPS());

  if(!(piSR[0]&1))//if(1) //
  {
	  UInt uiWidthMinPU      = g_uiMaxCUWidth>>g_uiMaxCUDepth;
	  UInt uiHeightMinPU     = g_uiMaxCUHeight>>g_uiMaxCUDepth;
    Int  iPUWidth = (piSR[0]<<3)&(~0xf);
    Int  iPUHeight = (piSR[1]<<3)&(~0xf);
	  UInt unitNum = max (1, (Int)((iPUWidth/uiWidthMinPU)*(iPUHeight/uiHeightMinPU)) ); 

    for ( UInt uiCUAddr = 0; uiCUAddr < pcPicDst->getPicSym()->getNumberOfCUsInFrame(); uiCUAddr++ )
    {
      TComDataCU *pcILRCU = pcPicDst->getPicSym()->getCU(uiCUAddr);
      UInt uiCUX = pcILRCU->getCUPelX();
      UInt uiCUY = pcILRCU->getCUPelY();
      TComMv cBlkSize;
      for(UInt uiAbsPartIdx=0; uiAbsPartIdx<pcPicDst->getNumPartInCU(); uiAbsPartIdx+=unitNum)
      {
        UInt uiPartX = uiCUX + g_auiRasterToPelX[g_auiZscanToRaster[uiAbsPartIdx]];
        UInt uiPartY = uiCUY + g_auiRasterToPelY[g_auiZscanToRaster[uiAbsPartIdx]];
        if(uiPartX >= uiPicWidth || uiPartY >= uiPicHeight)
          continue;
        Int iBlkWidth = min(iPUWidth, (Int)(uiPicWidth - uiPartX));
        Int iBlkHeight = min(iPUHeight, (Int)(uiPicHeight - uiPartY));
        cBlkSize.set(iBlkWidth, iBlkHeight);
        getEnhPUBlock(pcPicSrc, uiCUAddr, uiAbsPartIdx, pcPicDst, &cBlkSize, piSR); 
      }
    }
  }
  else
  {
    Int  iPUWidth = (piSR[0]<<3);
    Int  iPUHeight = (piSR[1]<<3);
    TComMv cBlkPos, cBlkSize;
    for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += iPUHeight)
    {
      Int iBlkHeight = min(iPUHeight, (Int)(uiPicHeight - iBlkY));        
      for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += iPUWidth)
      {
        Int iBlkWidth = min(iPUWidth, (Int)(uiPicWidth - iBlkX));        
        cBlkPos.set(iBlkX, iBlkY);
        cBlkSize.set(iBlkWidth, iBlkHeight);
        getEnhBlock(pcPicSrc, pcPicDst, &cBlkPos, &cBlkSize, piSR);      
      }
    }
  }

}

#if ILR_TM_ENH_WEIGHT_SIGNAL
Void TComPrediction::calcEnhBlkModeDiff(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight)
{
  //dst = up(BL) + (1-weight)*(MC(ELx)-MC(Up(BLx)));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
  iWeight = WEIGHT_SCALE - iWeight;

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i+=2)
    {
      pDst[i] = Clip(pSrc[i]+ ((iWeight*(pMCUpBLx[i]-DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i+1] = Clip(pSrc[i+1]+ ((iWeight*(pMCUpBLx[i+1]-DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
#if ILR_TM_ENH_CHROMA_ENABLE
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
#endif
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
#if ILR_TM_ENH_CHROMA_ENABLE
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
#if ILR_TM_ENH_CHROMA_ENABLE
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = Clip(pSrc[i] + ((iWeight*(pMCUpBLx[i] - DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = Clip(pSrc[i] + pMCUpBLx[i] - DIFFPIC_OFFSET);
    }
#else
    memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if ILR_TM_ENH_CHROMA_ENABLE
    pMCUpBLx += iStrideMCUpBLx;
#endif
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
#if ILR_TM_ENH_CHROMA_ENABLE
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
#if ILR_TM_ENH_CHROMA_ENABLE
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = Clip(pSrc[i] + ((iWeight*(pMCUpBLx[i] - DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = Clip(pSrc[i] + pMCUpBLx[i] - DIFFPIC_OFFSET);
    }
#else
    memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if ILR_TM_ENH_CHROMA_ENABLE
    pMCUpBLx += iStrideMCUpBLx;
#endif
  }
}

Void TComPrediction::calcEnhBlkModeDiff2(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight)
{
  //dst = MC(ELx) + weight*(up(BL)-MC(Up(BLx)));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
  Int iStrideMCELx = m_pEnhPred[1]->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
  Pel *pMCELx = m_pEnhPred[1]->getLumaAddr();
  iWeight = WEIGHT_SCALE - iWeight;

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i+=2)
    {
      pDst[i] = Clip(pMCELx[i]+ ((iWeight*(pSrc[i]-pMCELx[i]+pMCUpBLx[i]-DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i+1] = Clip(pMCELx[i+1]+ ((iWeight*(pSrc[i+1]-pMCELx[i+1]+pMCUpBLx[i+1]-DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
    pMCELx += iStrideMCELx;
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
  iStrideMCELx = m_pEnhPred[1]->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
#if ILR_TM_ENH_CHROMA_ENABLE
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
  //pMCELx = m_pEnhPred[1]->getCbAddr();
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
#if ILR_TM_ENH_CHROMA_ENABLE
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = Clip(pMCELx[i] + ((iWeight*(pSrc[i]-pMCELx[i]+pMCUpBLx[i] - DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = Clip(pSrc[i] + pMCUpBLx[i] - DIFFPIC_OFFSET);
    }
#else
    memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if ILR_TM_ENH_CHROMA_ENABLE
    pMCUpBLx += iStrideMCUpBLx;
    //pMCELx += iStrideMCELx;
#endif
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
#if ILR_TM_ENH_CHROMA_ENABLE
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
  //pMCELx = m_pEnhPred[1]->getCrAddr();
#endif
  for(Int j=0; j<iBlkHeight; j++)
  {
#if ILR_TM_ENH_CHROMA_ENABLE
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = Clip(pMCELx[i] + ((iWeight*(pSrc[i]-pMCELx[i]+pMCUpBLx[i] - DIFFPIC_OFFSET)+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = Clip(pSrc[i]+pMCUpBLx[i] - DIFFPIC_OFFSET);
    }
#else
    memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
    pDst += iStrideDst;
    pSrc += iStrideSrc;
#if ILR_TM_ENH_CHROMA_ENABLE
    pMCUpBLx += iStrideMCUpBLx;
    //pMCELx += iStrideMCELx;
#endif
  }
}

Void TComPrediction::calcEnhBlkMode1(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight)
{
  //dst = iWeight*up(BL) + (1-iWeight)(MC(ELx)));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
  Int iWeight1 = WEIGHT_SCALE - iWeight;
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i+=2)
    {
      pDst[i] = ((iWeight*pSrc[i] + iWeight1*pMCUpBLx[i] + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT);
      pDst[i+1] = ((iWeight*pSrc[i+1] + iWeight1*pMCUpBLx[i+1] + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
  }

  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ((iWeight*pSrc[i] + iWeight1*pMCUpBLx[i] + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT);
      pDst[i] = ((pSrc[i] + pMCUpBLx[i] + 1)>>1);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ((iWeight*pSrc[i] + iWeight1*pMCUpBLx[i] + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT);
      pDst[i] = ((pSrc[i] + pMCUpBLx[i] + 1)>>1);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
  }
}

Void TComPrediction::calcEnhBlkMode2(TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iIdx)
{
  //dst = (MC(ELx)));
  TComYuv *pEnhPred = m_pEnhPred[iIdx];
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = pEnhPred->getStride();
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = pEnhPred->getLumaAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    memcpy(pDst, pMCUpBLx, iBlkWidth*sizeof(Pel));
    pDst += iStrideDst;
    pMCUpBLx += iStrideMCUpBLx;
  }

  iStrideDst = pcPicYuvDst->getCStride();
  iStrideMCUpBLx = pEnhPred->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = pEnhPred->getCbAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    memcpy(pDst, pMCUpBLx, iBlkWidth*sizeof(Pel));
    pDst += iStrideDst;
    pMCUpBLx += iStrideMCUpBLx;
  }

  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = pEnhPred->getCrAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    memcpy(pDst, pMCUpBLx, iBlkWidth*sizeof(Pel));
    pDst += iStrideDst;
    pMCUpBLx += iStrideMCUpBLx;
  }
}

Void TComPrediction::getEnhBlockWeight(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize, Int *piSR
  )      
{
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
  TComDataCU *pcBLCU;
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  TComPicYuv *pcPicYuvDst = pcPicDst->getPicYuvRec();
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  Int iRefIdx0, iRefIdx1;
  
  pcPicBL->deriveUnitIdxBase(iBlkX /*+ (iBlkWidth>>1)*/, iBlkY /*+ (iBlkHeight>>1)*/, *piSR, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcPicBL->getCU(uiBaseCUAddr);
  iRefIdx0 = pcBLCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiBaseAbsPartIdx);
  iRefIdx1 = pcBLCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiBaseAbsPartIdx);  
  
  if(iRefIdx0 >=0 || iRefIdx1 >=0)
  {
    //TComDataCU *pcILRCU = pcPicDst->getPicSym()->getCU(0);
    Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
    Int iMode, iWeight;

    if(bi)
    {
      Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
      Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcBLCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiBaseAbsPartIdx) == pcBLCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiBaseAbsPartIdx))
      {
        bi = false;
      }
    }
    if(!bi)
    {
      //assert(iRefIdx0 < pcSliceBL->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSliceBL->getNumRefIdx(REF_PIC_LIST_1));
      RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
      Int        iRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;
      Int iWeightIdx = pcSlice->getWeightIdxUni(list, iRefIdx);

      iMode = pcSlice->getModeArr(0)[iWeightIdx];
      iWeight = pcSlice->getWeightArr(0)[iWeightIdx];
      //if(iMode < ILR_MIX_MODE)
      {
      TComMv     cMv     = pcBLCU->getCUMvField( list )->getMv( uiBaseAbsPartIdx );
      TComPicYuv *pcUpBLxRef;
      pcUpBLxRef = pcSlice->getRefPic( list, iRefIdx )->getFullPelBaseRec();

      cMv.scale(piSR);
      pcSlice->clipMv(cMv, iBlkX, iBlkY);
#if ILR_TM_ENH_MC_BILINEAR
      //Up(BLx); 
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#endif
#else
      //Up(BLx); 
      xPredInterLumaBlkByPos  ( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pEnhPred[0], bi );
#endif
#endif
      }
    }
    else
    {
      //assert(iRefIdx0 < pcSliceBL->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSliceBL->getNumRefIdx(REF_PIC_LIST_1));
      Int iWeightIdx = pcSlice->getWeightIdxBi(iRefIdx0, iRefIdx1);
      iMode = pcSlice->getModeArr(1)[iWeightIdx];
      iWeight = pcSlice->getWeightArr(1)[iWeightIdx];
      //if(iMode < ILR_MIX_MODE)
      {
      TComMv cMv = pcBLCU->getCUMvField( REF_PIC_LIST_0 )->getMv( uiBaseAbsPartIdx );
      TComMv cMv1 = pcBLCU->getCUMvField( REF_PIC_LIST_1 )->getMv( uiBaseAbsPartIdx );
      TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();

      cMv.scale(piSR);
      cMv1.scale(piSR);

      pcSlice->clipMv(cMv, iBlkX, iBlkY);
      pcSlice->clipMv(cMv1, iBlkX, iBlkY);
#if ILR_TM_ENH_MC_BILINEAR
      //Up(BLx); 
      TComYuv* pcMbYuv = m_acYuvPred;
      //TComPicYuv *pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
      xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE      
      xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      //Up(BLx); 
      TComYuv* pcMbYuv = m_acYuvPred;
      //TComPicYuv *pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
      xPredInterLumaBlkByPos  ( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef0, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#endif
      pcMbYuv = m_acYuvPred+1;
      //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
      xPredInterLumaBlkByPos  ( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
#if ILR_TM_ENH_CHROMA_ENABLE
      xPredInterChromaBlkByPos( pcUpBLxRef1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, pcMbYuv, bi );
      m_pEnhPred[0]->addAvg( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#else
      m_pEnhPred[0]->addAvgLuma( m_acYuvPred, m_acYuvPred+1, 0, iBlkWidth, iBlkHeight );
#endif
#endif
      }
    }
    //calc;
    if(iMode == ILR_DIFF_MODE)
    {
      if(iWeight)
        calcEnhBlkModeDiff(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight, iWeight);
      else
        calcEnhBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
    }
    else
    {
      //assert(!"Not supported yet!");
      printf("iMode:(%d) is not supported yet, treated as intra mode", iMode);
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);      
    }
#if 0
    else if(iMode == ILR_MIX_MODE)
      calcEnhBlkMode1(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight, iWeight);
    else if(iMode == ILR_EL_MODE)
      calcEnhBlkMode2(pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight, 1);
    else
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
#endif
  }
  else
  {
    //pcPicYuvUpBL -> pcPicYuvDst
    copyBlk(pcPicYuvUpBL, pcPicYuvDst, iBlkX, iBlkY, iBlkWidth, iBlkHeight);
  }
}

Void TComPrediction::enhPicWithTMWeight(TComPic *pcPicSrc, TComPic *pcPicDst, Int *piSR
    )
{
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  UInt uiPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  UInt uiPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();
  //set SPS;
  pcPicDst->getSlice(0)->setSPS(pcSlice->getSPS());

  Int  iPUWidth = (piSR[0]<<3);
  Int  iPUHeight = (piSR[1]<<3);
  TComMv cBlkPos, cBlkSize;
  for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += iPUHeight)
  {
    Int iBlkHeight = min(iPUHeight, (Int)(uiPicHeight - iBlkY));
    for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += iPUWidth)
    {
      Int iBlkWidth = min(iPUWidth, (Int)(uiPicWidth - iBlkX));              
      cBlkPos.set(iBlkX, iBlkY);
      cBlkSize.set(iBlkWidth, iBlkHeight);
      getEnhBlockWeight(pcPicSrc, pcPicDst, &cBlkPos, &cBlkSize, piSR);      
    }
  }
}
#endif

#endif

#if ILR_TM_ENH_COMPLEXITY
Void TComPrediction::xCalcMCMemoryAccessLumaByPos( Int iLumaPUPosX, Int iLumaPUPosY, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV, Int iDesignType)
{
  /* Derive reference block position and filter tap length for luma. */
  Int iLumaPosX = pcMv->getHor() >> 2;
  Int iLumaPosY = pcMv->getVer() >> 2;
  int ixFrac = (pcMv->getHor() & 0x3);
  int iyFrac = (pcMv->getVer() & 0x3);
  
  //this is only for DCT-IF;
  if(iFilTapH == NTAPS_LUMA)
  {
    if (ixFrac  == 1 || ixFrac  == 3) iFilTapH=iFilTapH-1;
    if (iyFrac  == 1 || iyFrac  == 3) iFilTapV=iFilTapV-1;
  }

  Int iMCTapH = (ixFrac != 0) ? iFilTapH : 0;
  Int iMCTapV = (iyFrac != 0) ? iFilTapV : 0;


  /* Calculate memory access bandwidth for luma. */
  if(iDesignType==3)
  {    
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalcPic_enhILR[i])
      {
        m_apcMemAccessCalcPic_enhILR[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
      }
    }
  }
  else if(iDesignType==2)
  {    
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalcPic[i])
      {
        m_apcMemAccessCalcPic[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
      }
    }
  }
  else if(iDesignType==1)
  {
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalcBlock[i])
      {
        m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
      }
    }
  }
  else
  {
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalc[i])
      {
        m_apcMemAccessCalc[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
      }
    }
  }
}
Void TComPrediction::xCalcMCMemoryAccessChromaByPos( Int iLumaPUPosX, Int iLumaPUPosY, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV, Int iDesignType)
{
  /* Derive reference block position and filter tap length for chroma. */
  Int iChromaPUPosX = iLumaPUPosX / 2;
  Int iChromaPUPosY = iLumaPUPosY / 2;
  Int iChromaPosX = pcMv->getHor() >> 3;
  Int iChromaPosY = pcMv->getVer() >> 3;
  Int iMCTapH = ((pcMv->getHor() & 0x7) != 0) ? iFilTapH : 0;
  Int iMCTapV = ((pcMv->getVer() & 0x7) != 0) ? iFilTapV : 0;

  /* Calculate memory access bandwidth for chroma. */
  if(iDesignType == 3)
  {
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalcPic_enhILR[i])
      {
        m_apcMemAccessCalcPic_enhILR[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
      }
    }
  }
  else if(iDesignType == 2)
  {
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalcPic[i])
      {
        m_apcMemAccessCalcPic[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
      }
    }
  }
  else if(iDesignType == 1)
  {
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalcBlock[i])
      {
        m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
      }
    }
  }
  else
  {
    for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
    {
      if (m_apcMemAccessCalc[i])
      {
        m_apcMemAccessCalc[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
      }
    }
  }
}

Void TComPrediction::xCalcUSMemoryAccessLumaByPos( Int iLumaPUPosX, Int iLumaPUPosY, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for luma. */
  Int iMCTapH = iFilTapH;
  Int iMCTapV = iFilTapV;

  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcBlock[i])
    {
		if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
		{
			m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW), int(iHeight*g_d_ScalingFactorH), int(iLumaPUPosX*g_d_ScalingFactorW), int(iLumaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, false);
		}
		else  // SNR case
		{
			m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW), int(iHeight*g_d_ScalingFactorH), int(iLumaPUPosX*g_d_ScalingFactorW), int(iLumaPUPosY*g_d_ScalingFactorH), 1, 1, false);
		}
    }
  }
}
Void TComPrediction::xCalcUSMemoryAccessChromaByPos( Int iLumaPUPosX, Int iLumaPUPosY, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for chroma. */
  Int iChromaPUPosX = iLumaPUPosX / 2;
  Int iChromaPUPosY = iLumaPUPosY / 2;
  
  Int iMCTapH = iFilTapH;
  Int iMCTapV = iFilTapV;

  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcBlock[i])
    {
		if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
		{
		      m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, true);
		}
		else // SNR case
		{
		      //m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, true);
		      m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), 1, 1, true);// jdong bug fix
		}
    }
  }
}

Void TComPrediction::getMCOperations(TComMv *pcMv, Int iBlkWidth, Int iBlkHeight, Int iLumaFilterTap, Int iChromaFilterTap, Int iDesignType)
{
  Int iMults=0, iAdds=0;
  //luma;
  Int iFracX = pcMv->getHor()&3;
  Int iFracY = pcMv->getVer()&3;
  Int iFilterTap = iLumaFilterTap;
  if( (!iFracY && iFracX) || (iFracY && !iFracX) )
  {
	  iMults +=iBlkWidth *iBlkHeight*iFilterTap; 
	  iAdds +=iBlkWidth *iBlkHeight*(iFilterTap-1);
  }
  else if(iFracX && iFracY)
  {
	  iMults +=iBlkWidth *(iBlkHeight+iFilterTap-1)*iFilterTap; 
	  iAdds +=iBlkWidth * (iBlkHeight+iFilterTap-1)*(iFilterTap-1);

    iMults +=iBlkWidth *iBlkHeight*iFilterTap; 
	  iAdds +=iBlkWidth *iBlkHeight*(iFilterTap-1);
  }

  //chroma;
  iFracX = pcMv->getHor()&7;
  iFracY = pcMv->getVer()&7;
  iFilterTap = iChromaFilterTap;
  iBlkHeight >>= 1;
  if( (!iFracY && iFracX) || (iFracY && !iFracX) )
  {
	  iMults +=iBlkWidth *iBlkHeight*iFilterTap; 
	  iAdds +=iBlkWidth *iBlkHeight*(iFilterTap-1);
  }
  else if(iFracX && iFracY)
  {
	  iMults +=iBlkWidth *(iBlkHeight +iFilterTap-1)*iFilterTap; 
	  iAdds +=iBlkWidth * (iBlkHeight +iFilterTap-1)*(iFilterTap-1);

    iMults +=iBlkWidth *iBlkHeight*iFilterTap; 
	  iAdds +=iBlkWidth *iBlkHeight*(iFilterTap-1);
  }

  if(iDesignType == 3)
  {
    g_d_num_mults_pic_enhILR += iMults;
    g_d_num_adds_pic_enhILR += iAdds;
  }
  else if(iDesignType == 2)
  {
    g_d_num_mults_pic += iMults;
    g_d_num_adds_pic += iAdds;
  }
  else if(iDesignType == 1)
  {
    g_d_num_mults_block += iMults;
    g_d_num_adds_block += iAdds;
  }
  else
  {
    g_d_num_mults += iMults;
    g_d_num_adds += iAdds;
  }
}

Void TComPrediction::getUSOperations(Int *piSR, Int *piWidth, Int *piHeight, Int iLumaTaps, Int iChromaTaps, Int iDesignType)
{
  if (piSR[0] != 2 || piSR[1] !=2 ) // !SNR case
  {
    Int iAdds=0, iMults=0;
    Int iWidth = *piWidth;
    Int iHeight = *piHeight;
	  // horizontal Luma interpolation
	  iMults += iWidth* (iHeight+iLumaTaps-1)* iLumaTaps; 
	  iAdds += iWidth* (iHeight+iLumaTaps-1)*(iLumaTaps-1);
	   // vertical Luma interpolation
	  iMults += iWidth* iHeight* iLumaTaps; 
	  iAdds += iWidth* iHeight*(iLumaTaps-1);

	   // horizontal Chroma interpolation
    Int iChromaWidth = *(piWidth+1); //iWidth>>1;
    Int iChromaHeight = *(piHeight+1); //iHeight>>1;
	  iMults += 2*iChromaWidth* (iChromaHeight + iChromaTaps-1)* iChromaTaps; 
	  iAdds += 2*iChromaWidth* (iChromaHeight + iChromaTaps-1)*(iChromaTaps-1);
	   // vertical Chroma interpolation
	  iMults +=2* iChromaWidth* iChromaHeight* iChromaTaps; 
	  iAdds += 2*iChromaWidth* iChromaHeight*(iChromaTaps-1);

    if(iDesignType==3)
    {
      g_d_num_mults_pic_enhILR += iMults;
      g_d_num_adds_pic_enhILR  += iAdds;
    }
    else if(iDesignType==2)
    {
      g_d_num_mults_pic += iMults;
      g_d_num_adds_pic  += iAdds;
    }
    else if(iDesignType==1)
    {
      g_d_num_mults_block += iMults;
      g_d_num_adds_block  += iAdds;
    }
    else
    {
      g_d_num_mults += iMults;
      g_d_num_adds  += iAdds;
    }
  }
}

Void TComPrediction::getMCBlockSize(TComMv *pcMv, Int iBlkWidth, Int iBlkHeight, Int iFilterTap, Int *piMCBlkWidth, Int *piMCBlkHeight)
{
  Int iFracX = pcMv->getHor()&3;
  Int iFracY = pcMv->getVer()&3;
  
  piMCBlkWidth[0] = iFracX? iBlkWidth+iFilterTap-1 : iBlkWidth ;
  piMCBlkHeight[0] = iFracY? iBlkHeight+iFilterTap-1 : iBlkHeight;
 
  //chroma;
  iFracX = pcMv->getHor()&7;
  iFracY = pcMv->getVer()&7;
  piMCBlkWidth[1] = iFracX? iBlkWidth/2+iFilterTap-1 : iBlkWidth/2 ;
  piMCBlkHeight[1] = iFracY? iBlkHeight/2+iFilterTap-1 : iBlkHeight/2;
}

Void TComPrediction::calcCompEnhILRBlock(TComPic *pcPicSrc, TComMv *pcBlkPos, TComMv *pcBlkSize, Int *piSR, Int iDesignType)      
{
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
  TComDataCU *pcBLCU;
  //TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  Int iRefIdx0, iRefIdx1;

  pcPicBL->deriveUnitIdxBase(iBlkX /*+ (iBlkWidth>>1)*/, iBlkY /*+ (iBlkHeight>>1)*/, *piSR, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcPicBL->getCU(uiBaseCUAddr);
  iRefIdx0 = pcBLCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiBaseAbsPartIdx);
  iRefIdx1 = pcBLCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiBaseAbsPartIdx);  
  
  if(iRefIdx0 >=0 || iRefIdx1 >=0)
  {
    //TComDataCU *pcILRCU = pcPicDst->getPicSym()->getCU(0);
    Int iWeight;
    Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
    if(bi)
    {      
      Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
      Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcBLCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiBaseAbsPartIdx) == pcBLCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiBaseAbsPartIdx))
      {
        bi = false;
      }
    }
    if(!bi)
    {
      //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
      RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
      Int        iRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;
      Int iWeightIdx = pcSlice->getWeightIdxUni(list, iRefIdx);
      //Int iMode = pcSlice->getModeArr(0)[iWeightIdx];
      iWeight = pcSlice->getWeightArr(0)[iWeightIdx];

      TComMv     cMv     = pcBLCU->getCUMvField( list )->getMv( uiBaseAbsPartIdx );
      cMv.scale(piSR);
      pcSlice->clipMv(cMv, iBlkX, iBlkY);
      
      if(iDesignType == 3 )
      {
#if 1
        //use D picture to generate the residual;
        initMCMemoryAccessCalculator( pcSlice->getRefPic( list, iRefIdx ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, list, iRefIdx, &cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, list, iRefIdx, &cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        getMCOperations(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
#else
        //use ELx and Up(BLx) pictures to generate the residual;
        TComMv zeroMV(0,0);
        Int iMCBlkWidth[2], iMCBlkHeight[2];
        getMCBlockSize(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, iMCBlkWidth, iMCBlkHeight);
        Int iMCBlkX = iBlkX + (cMv.getHor()>>2) - (cMv.getHor()? ((NTAPS_BILINEAR-1)>>1): 0);
        Int iMCBlkY = iBlkY + (cMv.getVer()>>2) - (cMv.getVer()? ((NTAPS_BILINEAR-1)>>1): 0);
        //fetch MC block from ELx;
        initMCMemoryAccessCalculator( pcSlice->getRefPic( list, iRefIdx ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iMCBlkX, iMCBlkY, list, iRefIdx, &zeroMV, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iMCBlkX, iMCBlkY, list, iRefIdx, &zeroMV, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);
        //fetch MC block from Up(BLx);
        xCalcMCMemoryAccessLumaByPos  ( iMCBlkX, iMCBlkY, list, iRefIdx, &zeroMV, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iMCBlkX, iMCBlkY, list, iRefIdx, &zeroMV, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);
        //subtract, offset;
        g_d_num_adds_pic_enhILR += (iMCBlkWidth[0]*iMCBlkHeight[0] + 2*iMCBlkWidth[1]*iMCBlkHeight[1])*2;
        getMCOperations(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
#endif
      }
      else if(iDesignType == 1 )
      {
        Int iMCBlkWidth[2], iMCBlkHeight[2];
        getMCBlockSize(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, iMCBlkWidth, iMCBlkHeight);
        //upsample to iMCBlkWidth x iMCBlkHeight;	      
        Int iMCBlkX = iBlkX + (cMv.getHor()>>2) - (cMv.getHor()? ((NTAPS_BILINEAR-1)>>1): 0);
        Int iMCBlkY = iBlkY + (cMv.getVer()>>2) - (cMv.getVer()? ((NTAPS_BILINEAR-1)>>1): 0);
        initMCMemoryAccessCalculator(pcSliceBL->getRefPic(list, iRefIdx), iDesignType); // for block-base up-sampling
        xCalcUSMemoryAccessLumaByPos(iMCBlkX, iMCBlkY, iMCBlkWidth[0], iMCBlkHeight[0], LUMA_USF_TAPS, LUMA_USF_TAPS);
        xCalcUSMemoryAccessChromaByPos(iMCBlkX, iMCBlkY, (iMCBlkWidth[1]<<1), (iMCBlkHeight[1]<<1), CHROMA_USF_TAPS, CHROMA_USF_TAPS);
        getUSOperations(piSR, iMCBlkWidth, iMCBlkHeight, LUMA_USF_TAPS, CHROMA_USF_TAPS, iDesignType);
        //fetch EL block;
        TComMv tmpMv;
        tmpMv.set(((iMCBlkX-iBlkX)>>1)<<3, ((iMCBlkY-iBlkY)>>1)<<3); 
        initMCMemoryAccessCalculator(pcSlice->getRefPic( list, iRefIdx ), iDesignType); // for block-base up-sampling
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, list, iRefIdx, &tmpMv, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, list, iRefIdx, &tmpMv, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);

        //subtract, offset;
        g_d_num_adds_block += (iMCBlkWidth[0]*iMCBlkHeight[0] + 2*iMCBlkWidth[1]*iMCBlkHeight[1])*2;

        //interpolate;
        /*
        tmpMv.set(cMv.getHor()&3, cMv.getVer()&3); 
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, list, iRefIdx, &tmpMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        tmpMv.set(cMv.getHor()&7, cMv.getVer()&7); 
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, list, iRefIdx, &tmpMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        */
        getMCOperations(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
      }
    }
    else
    {
      //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
      //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
      Int iWeightIdx = pcSlice->getWeightIdxBi(iRefIdx0, iRefIdx1);
      //Int iMode = pcSlice->getModeArr(1)[iWeightIdx];
      iWeight = pcSlice->getWeightArr(1)[iWeightIdx];

      TComMv cMv = pcBLCU->getCUMvField( REF_PIC_LIST_0 )->getMv( uiBaseAbsPartIdx );
      TComMv cMv1 = pcBLCU->getCUMvField( REF_PIC_LIST_1 )->getMv( uiBaseAbsPartIdx );

      cMv.scale(piSR);
      cMv1.scale(piSR);
      pcSlice->clipMv(cMv, iBlkX, iBlkY);
      pcSlice->clipMv(cMv1, iBlkX, iBlkY);

      if(iDesignType == 3 )
      {
#if 1
        //use D picture to generate the residual;
        initMCMemoryAccessCalculator( pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_0, iRefIdx0, &cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_0, iRefIdx0, &cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        getMCOperations(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);

        initMCMemoryAccessCalculator( pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_1, iRefIdx1, &cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_1, iRefIdx1, &cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        getMCOperations(&cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);

        //average;
        g_d_num_adds_pic_enhILR += iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2); 
#else
        //use ELx and Up(BLx) pictures to generate the residual;
        TComMv zeroMV(0,0);
        Int iMCBlkWidth[2], iMCBlkHeight[2];
        getMCBlockSize(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, iMCBlkWidth, iMCBlkHeight);
        Int iMCBlkX = iBlkX + (cMv.getHor()>>2) - (cMv.getHor()? ((NTAPS_BILINEAR-1)>>1): 0);
        Int iMCBlkY = iBlkY + (cMv.getVer()>>2) - (cMv.getVer()? ((NTAPS_BILINEAR-1)>>1): 0);
        //fetch MC block from ELx;
        initMCMemoryAccessCalculator( pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iMCBlkX, iMCBlkY, REF_PIC_LIST_0, iRefIdx0, &zeroMV, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iMCBlkX, iMCBlkY, REF_PIC_LIST_0, iRefIdx0, &zeroMV, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);
        //fetch MC block from Up(BLx);
        initMCMemoryAccessCalculator( pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iMCBlkX, iMCBlkY, REF_PIC_LIST_0, iRefIdx0, &zeroMV, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iMCBlkX, iMCBlkY, REF_PIC_LIST_0, iRefIdx0, &zeroMV, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);
        //subtract, offset;
        g_d_num_adds_pic_enhILR += (iMCBlkWidth[0]*iMCBlkHeight[0] + 2*iMCBlkWidth[1]*iMCBlkHeight[1])*2;
        getMCOperations(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);

        //cMv1;
        //use ELx and Up(BLx) pictures to generate the residual;
        getMCBlockSize(&cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, iMCBlkWidth, iMCBlkHeight);
        iMCBlkX = iBlkX + (cMv1.getHor()>>2) - (cMv1.getHor()? ((NTAPS_BILINEAR-1)>>1): 0);
        iMCBlkY = iBlkY + (cMv1.getVer()>>2) - (cMv1.getVer()? ((NTAPS_BILINEAR-1)>>1): 0);
        //fetch MC block from ELx;
        initMCMemoryAccessCalculator( pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iMCBlkX, iMCBlkY, REF_PIC_LIST_1, iRefIdx1, &zeroMV, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iMCBlkX, iMCBlkY, REF_PIC_LIST_1, iRefIdx1, &zeroMV, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);
        //fetch MC block from Up(BLx);
        initMCMemoryAccessCalculator( pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 ), iDesignType );
        xCalcMCMemoryAccessLumaByPos  ( iMCBlkX, iMCBlkY, REF_PIC_LIST_1, iRefIdx1, &zeroMV, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iMCBlkX, iMCBlkY, REF_PIC_LIST_1, iRefIdx1, &zeroMV, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);
        //subtract, offset;
        g_d_num_adds_pic_enhILR += (iMCBlkWidth[0]*iMCBlkHeight[0] + 2*iMCBlkWidth[1]*iMCBlkHeight[1])*2;
        getMCOperations(&cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);

        //average;
        g_d_num_adds_pic_enhILR += iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2); 
#endif
      }
      else if(iDesignType == 1 )
      {
        Int iMCBlkWidth[2], iMCBlkHeight[2];
        getMCBlockSize(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, iMCBlkWidth, iMCBlkHeight);
        //upsample to iMCBlkWidth x iMCBlkHeight;
        Int iMCBlkX = iBlkX + (cMv.getHor()>>2) - (cMv.getHor()? ((NTAPS_BILINEAR-1)>>1): 0);
        Int iMCBlkY = iBlkY + (cMv.getVer()>>2) - (cMv.getVer()? ((NTAPS_BILINEAR-1)>>1): 0);
        initMCMemoryAccessCalculator( pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0), iDesignType );
        xCalcUSMemoryAccessLumaByPos(iMCBlkX, iMCBlkY, iMCBlkWidth[0], iMCBlkHeight[0], LUMA_USF_TAPS, LUMA_USF_TAPS);
        xCalcUSMemoryAccessChromaByPos(iMCBlkX, iMCBlkY, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, CHROMA_USF_TAPS, CHROMA_USF_TAPS);
        getUSOperations(piSR, iMCBlkWidth, iMCBlkHeight, LUMA_USF_TAPS, CHROMA_USF_TAPS, iDesignType);
        //fetch EL block;
        TComMv tmpMv;
        tmpMv.set(((iMCBlkX-iBlkX)>>1)<<3, ((iMCBlkY-iBlkY)>>1)<<3); 
        initMCMemoryAccessCalculator(pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 ), iDesignType); // for block-base up-sampling
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_0, iRefIdx0, &tmpMv, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_0, iRefIdx0, &tmpMv, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);

        //subtract, offset;
        g_d_num_adds_block += (iMCBlkWidth[0]*iMCBlkHeight[0] + 2*iMCBlkWidth[1]*iMCBlkHeight[1])*2;

        //interpolate;
        /*
        tmpMv.set(cMv.getHor()&3, cMv.getVer()&3); 
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_0, iRefIdx0, &tmpMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        tmpMv.set(cMv.getHor()&7, cMv.getVer()&7); 
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_0, iRefIdx0, &tmpMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        */
        getMCOperations(&cMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);

        //cMv1;
        getMCBlockSize(&cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, iMCBlkWidth, iMCBlkHeight);
        //upsample to iMCBlkWidth x iMCBlkHeight;
        iMCBlkX = iBlkX + (cMv1.getHor()>>2) - (cMv1.getHor()? ((NTAPS_BILINEAR-1)>>1): 0);
        iMCBlkY = iBlkY + (cMv1.getVer()>>2) - (cMv1.getVer()? ((NTAPS_BILINEAR-1)>>1): 0);
        initMCMemoryAccessCalculator( pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1), iDesignType );
        xCalcUSMemoryAccessLumaByPos(iMCBlkX, iMCBlkY, iMCBlkWidth[0], iMCBlkHeight[0], LUMA_USF_TAPS, LUMA_USF_TAPS);
        xCalcUSMemoryAccessChromaByPos(iMCBlkX, iMCBlkY, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, CHROMA_USF_TAPS, CHROMA_USF_TAPS);
        getUSOperations(piSR, iMCBlkWidth, iMCBlkHeight, LUMA_USF_TAPS, CHROMA_USF_TAPS, iDesignType);
        //fetch EL block;
        tmpMv.set(((iMCBlkX-iBlkX)>>1)<<3, ((iMCBlkY-iBlkY)>>1)<<3); 
        initMCMemoryAccessCalculator(pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 ), iDesignType); // for block-base up-sampling
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_1, iRefIdx1, &tmpMv, iMCBlkWidth[0], iMCBlkHeight[0], 1, 1, iDesignType);
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_1, iRefIdx1, &tmpMv, iMCBlkWidth[1]<<1, iMCBlkHeight[1]<<1, 1, 1, iDesignType);

        //subtract, offset;
        g_d_num_adds_block += (iMCBlkWidth[0]*iMCBlkHeight[0] + 2*iMCBlkWidth[1]*iMCBlkHeight[1])*2;

        //interpolate;
        /*
        tmpMv.set(cMv1.getHor()&3, cMv1.getVer()&3); 
        xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_1, iRefIdx1, &tmpMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        tmpMv.set(cMv1.getHor()&7, cMv1.getVer()&7); 
        xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_1, iRefIdx1, &tmpMv, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);
        */
        getMCOperations(&cMv1, iBlkWidth, iBlkHeight, NTAPS_BILINEAR, NTAPS_BILINEAR, iDesignType);

        //average;
        g_d_num_adds_block += iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2); 
      }
    }
    if(iDesignType==3)
    {
	    if(iWeight)
        g_d_num_mults_pic_enhILR += iBlkWidth*iBlkHeight; // + iBlkWidth*(iBlkHeight/2); 
			g_d_num_adds_pic_enhILR += (iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2))*2;  //offset, + 
    }
    else if(iDesignType==2)
    {
	    if(iWeight)
        g_d_num_mults_pic += iBlkWidth*iBlkHeight; // + iBlkWidth*(iBlkHeight/2); 
			g_d_num_adds_pic += (iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2))*2;  //offset, + 
    }
    else if(iDesignType==1)
    {
	    if(iWeight)
        g_d_num_mults_block+= iBlkWidth*iBlkHeight;// + iBlkWidth*(iBlkHeight/2); 
			g_d_num_adds_block += (iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2))*2; 
    }
    else
    {
	    if(iWeight)
        g_d_num_mults += iBlkWidth*iBlkHeight; // + iBlkWidth*(iBlkHeight/2); 
			g_d_num_adds += (iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2))*2; 
    }
  }
#if 0
  else if(iDesignType==3)
  { //PU based implementation is already there; Picture based method is already copied during the upsampling process;
    TComMv     cMv(0,0);
    initMCMemoryAccessCalculator( pcPicSrc, iDesignType );
    xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, (RefPicList)0, 0, &cMv, iBlkWidth, iBlkHeight, 1, 1, iDesignType);
    xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, (RefPicList)0, 0, &cMv, iBlkWidth, iBlkHeight, 1, 1, iDesignType);
  }
#endif
}

Int TComPrediction::sameMotion( TComPic *pPic, Int iBlkX0, Int iBlkY0, Int iBlkX1, Int iBlkY1)
{
  Int bSame = 0;
  TComSlice *pcSlice = pPic->getSlice(pPic->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
  TComDataCU *pcBLCU;
  UInt uiBaseCUAddr, uiBaseAbsPartIdx0, uiBaseAbsPartIdx1;
  Int iRefIdx0, iRefIdx1;
  Bool bBiPred0, bBiPred1, bInter0, bInter1;

  pcPicBL->deriveUnitIdxBase(iBlkX0, iBlkY0, g_iSRNum[0], uiBaseCUAddr, uiBaseAbsPartIdx0);
  pcBLCU = pcPicBL->getCU(uiBaseCUAddr);
  TComCUMvField *pMvField[4];
  pMvField[0] = pcBLCU->getCUMvField(REF_PIC_LIST_0);
  pMvField[1] = pcBLCU->getCUMvField(REF_PIC_LIST_1);
  pcPicBL->deriveUnitIdxBase(iBlkX1, iBlkY1, g_iSRNum[0], uiBaseCUAddr, uiBaseAbsPartIdx1);
  pcBLCU = pcPicBL->getCU(uiBaseCUAddr);
  pMvField[2] = pcBLCU->getCUMvField(REF_PIC_LIST_0);
  pMvField[3] = pcBLCU->getCUMvField(REF_PIC_LIST_1);

  iRefIdx0 = pMvField[0]->getRefIdx(uiBaseAbsPartIdx0);
  iRefIdx1 = pMvField[1]->getRefIdx(uiBaseAbsPartIdx0);
  bBiPred0 = (iRefIdx0 >=0 && iRefIdx1 >=0);
  bInter0 = (iRefIdx0 >=0 || iRefIdx1 >=0);
  if(bBiPred0)
  {      
    Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
    Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
    if(RefPOCL0 == RefPOCL1 && pMvField[0]->getMv(uiBaseAbsPartIdx0) == pMvField[1]->getMv(uiBaseAbsPartIdx0))
    {
      bBiPred0 = false;
    }
  }

  iRefIdx0 = pMvField[2]->getRefIdx(uiBaseAbsPartIdx1);
  iRefIdx1 = pMvField[3]->getRefIdx(uiBaseAbsPartIdx1);
  bBiPred1 = (iRefIdx0 >=0 && iRefIdx1 >=0);
  bInter1 = (iRefIdx0 >=0 || iRefIdx1 >=0);
  if(bBiPred1)
  {      
    Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
    Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
    if(RefPOCL0 == RefPOCL1 && pMvField[2]->getMv(uiBaseAbsPartIdx1) == pMvField[3]->getMv(uiBaseAbsPartIdx1))
    {
      bBiPred1 = false;
    }
  }
  if(bInter0 && bInter1)
  {
    if(bBiPred0 == bBiPred1)
    {
      if(bBiPred0)
      {
        Int iBlk0RefPOC0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, pMvField[0]->getRefIdx(uiBaseAbsPartIdx0))->getPOC();
        Int iBlk0RefPOC1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, pMvField[1]->getRefIdx(uiBaseAbsPartIdx0))->getPOC();
        Int iBlk1RefPOC0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, pMvField[2]->getRefIdx(uiBaseAbsPartIdx1))->getPOC();
        Int iBlk1RefPOC1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, pMvField[3]->getRefIdx(uiBaseAbsPartIdx1))->getPOC();
        if( (iBlk0RefPOC0 == iBlk1RefPOC0 && pMvField[0]->getMv(uiBaseAbsPartIdx0) == pMvField[2]->getMv(uiBaseAbsPartIdx1)) &&
            (iBlk0RefPOC1 == iBlk1RefPOC1 && pMvField[1]->getMv(uiBaseAbsPartIdx0) == pMvField[3]->getMv(uiBaseAbsPartIdx1)) )
        {
          bSame = 1;
        }
      }
      else
      {
        Int iBlk0List = pMvField[0]->getRefIdx(uiBaseAbsPartIdx0)>=0? 0: 1;
        Int iBlk1List = pMvField[2]->getRefIdx(uiBaseAbsPartIdx1)>=0? 0: 1;
        Int iBlk0RefPOC = pcSliceBL->getRefPic((RefPicList)iBlk0List, pMvField[iBlk0List]->getRefIdx(uiBaseAbsPartIdx0))->getPOC();
        Int iBlk1RefPOC = pcSliceBL->getRefPic((RefPicList)iBlk1List, pMvField[2+iBlk1List]->getRefIdx(uiBaseAbsPartIdx1))->getPOC();
        if(iBlk0RefPOC == iBlk1RefPOC && pMvField[iBlk0List]->getMv(uiBaseAbsPartIdx0) == pMvField[2+iBlk1List]->getMv(uiBaseAbsPartIdx1))
        {
          bSame = 1;
        }
      }
    }
  }
  else if(!bInter0 && !bInter1)
  {
    bSame = 1;
  }

  return bSame;
}

Void TComPrediction::calcEnhILRReconPic(TComDataCU *pcCU)
{
  TComSlice *pcSlice = pcCU->getSlice();
  UInt uiPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  UInt uiPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();
  TComPic *pCurrPic = pcCU->getPic();
  Int iDesignType = 3; //picture level operations;
#if 0
  if(pcSlice->getEnhILREnabled())
  {
    Int  iPUWidth = (g_iSRNum[0]<<3);
    Int  iPUHeight = (g_iSRNum[1]<<3);
    TComMv cBlkPos, cBlkSize;
    for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += iPUHeight)
    {
      Int iBlkHeight = min(iPUHeight, (Int)(uiPicHeight - iBlkY));        
      for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += iPUWidth)
      {
        Int iBlkWidth = min(iPUWidth, (Int)(uiPicWidth - iBlkX));        
        cBlkPos.set(iBlkX, iBlkY);
        cBlkSize.set(iBlkWidth, iBlkHeight);
        calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
      }
    }
  }
#else
  if(pcSlice->getEnhILREnabled())
  {
    Int  iPUWidth = (g_iSRNum[0]<<3);
    Int  iPUHeight = (g_iSRNum[1]<<3);
    Int  iPUWidthx2 = (g_iSRNum[0]<<4);
    Int  iPUHeightx2 = (g_iSRNum[1]<<4);
    TComMv cBlkPos, cBlkSize;
    for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += iPUHeightx2)
    {
      Int iBlkHeight = min(iPUHeightx2, (Int)(uiPicHeight - iBlkY));        
      Int iBdyY = iBlkY + iPUHeight;

      for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += iPUWidthx2)
      {
        Int iBlkWidth = min(iPUWidthx2, (Int)(uiPicWidth - iBlkX));        
        Int iBdyX = iBlkX + iPUWidth;

        if(iBlkX+iBlkWidth <=iBdyX && iBlkY+iBlkHeight<=iBdyY)
        {
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBlkWidth, iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
        else if(iBlkX+iBlkWidth <= iBdyX)
        { //to check if it is 2NxN case;
          Int iSubBlkX[2], iSubBlkY[2];        
          iSubBlkX[0] = iBlkX;
          iSubBlkY[0] = iBlkY;
          iSubBlkX[1] = iBlkX;
          iSubBlkY[1] = iBdyY;
          if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1]))
          {
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBlkWidth, iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else
          {
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBlkWidth, iBdyY - iBlkY);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

            cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
            cBlkSize.set(iBlkWidth, iBlkY + iBlkHeight - iSubBlkY[1]);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
        }
        else if(iBlkY+iBlkHeight<=iBdyY)
        { //to check if it is Nx2N case;
          Int iSubBlkX[2], iSubBlkY[2];        
          iSubBlkX[0] = iBlkX;
          iSubBlkY[0] = iBlkY;
          iSubBlkX[1] = iBdyX;
          iSubBlkY[1] = iBlkY;
          if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1]))
          {
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBlkWidth, iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else
          {
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBdyX - iBlkX, iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

            cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
            cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }       
        }
        else
        {
          Int iSubBlkX[4], iSubBlkY[4];        
          iSubBlkX[0] = iBlkX;
          iSubBlkY[0] = iBlkY;
          iSubBlkX[1] = iBdyX;
          iSubBlkY[1] = iBlkY;
          iSubBlkX[2] = iBlkX;
          iSubBlkY[2] = iBdyY;
          iSubBlkX[3] = iBdyX;
          iSubBlkY[3] = iBdyY;
          if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1])&& 
             sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[2], iSubBlkY[2])&&
             sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[3], iSubBlkY[3]))
          { //2Nx2N;
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBlkWidth, iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1])&& 
                  sameMotion(pCurrPic, iSubBlkX[2], iSubBlkY[2], iSubBlkX[3], iSubBlkY[3]))
          { //2NxN;
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBlkWidth, iBdyY - iBlkY);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

            cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
            cBlkSize.set(iBlkWidth, iBlkY + iBlkHeight - iSubBlkY[2]);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[2], iSubBlkY[2])&& 
                  sameMotion(pCurrPic, iSubBlkX[1], iSubBlkY[1], iSubBlkX[3], iSubBlkY[3]))
          { //Nx2N;
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBdyX - iBlkX, iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

            cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
            cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else
          {
            if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1]))
            {//subblock0, subblock1;
              cBlkPos.set(iBlkX, iBlkY);
              cBlkSize.set(iBlkWidth, iBdyY - iBlkY);
              calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              //subblock2;
              cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
              cBlkSize.set(iBdyX-iSubBlkX[2], iBlkY + iBlkHeight - iSubBlkY[2]);
              calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              //subblock3;
              cBlkPos.set(iSubBlkX[3], iSubBlkY[3]);
              cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[3], iBlkY + iBlkHeight - iSubBlkY[3]);
              calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
            }
            else if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[2], iSubBlkY[2]))
            {//subblock0, subblock2;
              cBlkPos.set(iBlkX, iBlkY);
              cBlkSize.set(iBdyX - iBlkX, iBlkHeight);
              calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              //subblock1;
              cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
              cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBdyY - iSubBlkY[1]);
              calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              //subblock3;
              cBlkPos.set(iSubBlkX[3], iSubBlkY[3]);
              cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[3], iBlkY + iBlkHeight - iSubBlkY[3]);
              calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
            }
            else
            {
              //subblock2, subblock3;
              if(sameMotion(pCurrPic, iSubBlkX[1], iSubBlkY[1], iSubBlkX[3], iSubBlkY[3]))
              {
                //subblock0;
                cBlkPos.set(iBlkX, iBlkY);
                cBlkSize.set(iBdyX - iBlkX, iBdyY - iBlkY);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock2;
                cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
                cBlkSize.set(iBdyX-iSubBlkX[2], iBlkY + iBlkHeight - iSubBlkY[2]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock1/3;
                cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBlkHeight);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              }
              else if(sameMotion(pCurrPic, iSubBlkX[2], iSubBlkY[2], iSubBlkX[3], iSubBlkY[3]))
              {
                //subblock0;
                cBlkPos.set(iBlkX, iBlkY);
                cBlkSize.set(iBdyX - iBlkX, iBdyY - iBlkY);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock1;
                cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBdyY - iSubBlkY[1]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock2/3;
                cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
                cBlkSize.set(iBlkWidth, iBlkY + iBlkHeight - iSubBlkY[2]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              }
              else
              {
                //subblock0;
                cBlkPos.set(iBlkX, iBlkY);
                cBlkSize.set(iBdyX - iBlkX, iBdyY - iBlkY);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock1;
                cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBdyY - iSubBlkY[1]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock2;
                cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
                cBlkSize.set(iBdyX-iSubBlkX[2], iBlkY + iBlkHeight - iSubBlkY[2]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock3;
                cBlkPos.set(iSubBlkX[3], iSubBlkY[3]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[3], iBlkY + iBlkHeight - iSubBlkY[3]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              }
            }
          }
        } //end partition check;

      }
    }
  }
#endif

#if 1
  //generate D;
  TComMv zeroMV(0,0);
  for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += g_uiMaxCUHeight)
  {
    Int iBlkHeight = min((Int)g_uiMaxCUHeight, (Int)(uiPicHeight - iBlkY));        
    for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += g_uiMaxCUWidth)
    {
      Int iBlkWidth = min((Int)g_uiMaxCUWidth, (Int)(uiPicWidth - iBlkX));        

      initMCMemoryAccessCalculator( pCurrPic, iDesignType );
      //reconstructed picture;
      xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_0, 0, &zeroMV, iBlkWidth, iBlkHeight, 1, 1, iDesignType);
      xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_0, 0, &zeroMV, iBlkWidth, iBlkHeight, 1, 1, iDesignType);
      //upsampled picture;
      initMCMemoryAccessCalculator( pCurrPic, iDesignType );
      xCalcMCMemoryAccessLumaByPos  ( iBlkX, iBlkY, REF_PIC_LIST_0, 0, &zeroMV, iBlkWidth, iBlkHeight, 1, 1, iDesignType);
      xCalcMCMemoryAccessChromaByPos( iBlkX, iBlkY, REF_PIC_LIST_0, 0, &zeroMV, iBlkWidth, iBlkHeight, 1, 1, iDesignType);
      
      //subtract, shift;
      g_d_num_adds_pic_enhILR += (iBlkWidth*iBlkHeight + iBlkWidth*(iBlkHeight/2))*2;    
    }
  }  
#endif
}

#if 0
Void TComPrediction::calcEnhILRReconBlock(TComDataCU *pcCU, UInt uiPartAddr, Int iWidth, Int iHeight)
{
  TComPic *pCurrPic = pcCU->getPic();
  Int iPUPosX = pCurrPic->getPicYuvRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int iPUPosY = pCurrPic->getPicYuvRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int  iMinPUWidth = (g_iSRNum[0]<<3);
  Int  iMinPUHeight = (g_iSRNum[1]<<3);
  TComMv cBlkPos, cBlkSize;
  Int iX0, iX1, iY0, iY1;
  Int iDesignType = 1; //block level operations;

  iX0 = iPUPosX/iMinPUWidth*iMinPUWidth;
  iX1 = iPUPosX+iWidth; 
  iY0 = iPUPosY/iMinPUHeight*iMinPUHeight;
  iY1 = iPUPosY+iHeight; 

  for ( Int iY=iY0; iY<iY1; iY += iMinPUHeight)
  {
    Int iBlkY = max(iY, iPUPosY);
    Int iBlkHeight = min(iMinPUHeight, (iY1 - iBlkY));        
    for(Int iX=iX0; iX<iX1; iX += iMinPUWidth)
    {
      Int iBlkX = max(iX, iPUPosX);              
      Int iBlkWidth = min(iMinPUWidth, (iX1 - iBlkX));
      
      cBlkPos.set(iBlkX, iBlkY);
      cBlkSize.set(iBlkWidth, iBlkHeight);
      calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
    }
  }
}
#else
//get the 2x2 process;
Void TComPrediction::calcEnhILRReconBlock(TComDataCU *pcCU, UInt uiPartAddr, Int iWidth, Int iHeight)
{
  TComPic *pCurrPic = pcCU->getPic();
  Int iPUPosX = pCurrPic->getPicYuvRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int iPUPosY = pCurrPic->getPicYuvRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int  iMinPUWidth = (g_iSRNum[0]<<3);
  Int  iMinPUWidthx2 = (g_iSRNum[0]<<4);
  Int  iMinPUHeight = (g_iSRNum[1]<<3);
  Int  iMinPUHeightx2 = (g_iSRNum[1]<<4);
  TComMv cBlkPos, cBlkSize;
  Int iX0, iX1, iY0, iY1;
  Int iDesignType = 1; //block level operations;

  iX0 = iPUPosX/iMinPUWidth*iMinPUWidth;
  iX1 = iPUPosX+iWidth; 
  iY0 = iPUPosY/iMinPUHeight*iMinPUHeight;
  iY1 = iPUPosY+iHeight; 

  for ( Int iY=iY0; iY<iY1; iY += iMinPUHeightx2)
  {
    Int iBlkY = max(iY, iPUPosY);
    Int iBlkHeight = min(iMinPUHeightx2, (iY1 - iBlkY));        
    for(Int iX=iX0; iX<iX1; iX += iMinPUWidthx2)
    {
      Int iBlkX = max(iX, iPUPosX);              
      Int iBlkWidth = min(iMinPUWidthx2, (iX1 - iBlkX));
      
      //check the partition;
      Int iBdyX = iX+iMinPUWidth;
      Int iBdyY = iY+iMinPUHeight;
      if(iBlkX+iBlkWidth <= iBdyX && iBlkY+iBlkHeight<=iBdyY)
      {
        cBlkPos.set(iBlkX, iBlkY);
        cBlkSize.set(iBlkWidth, iBlkHeight);
        calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
      }
      else if(iBlkX+iBlkWidth <= iBdyX)
      { //to check if it is 2NxN case;
        Int iSubBlkX[2], iSubBlkY[2];        
        iSubBlkX[0] = iBlkX;
        iSubBlkY[0] = iBlkY;
        iSubBlkX[1] = iBlkX;
        iSubBlkY[1] = iBdyY;
        if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1]))
        {
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBlkWidth, iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
        else
        {
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBlkWidth, iBdyY - iBlkY);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

          cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
          cBlkSize.set(iBlkWidth, iBlkY + iBlkHeight - iSubBlkY[1]);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
      }
      else if(iBlkY+iBlkHeight<=iBdyY)
      { //to check if it is Nx2N case;
        Int iSubBlkX[2], iSubBlkY[2];        
        iSubBlkX[0] = iBlkX;
        iSubBlkY[0] = iBlkY;
        iSubBlkX[1] = iBdyX;
        iSubBlkY[1] = iBlkY;
        if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1]))
        {
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBlkWidth, iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
        else
        {
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBdyX - iBlkX, iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

          cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
          cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }       
      }
      else
      {
        Int iSubBlkX[4], iSubBlkY[4];        
        iSubBlkX[0] = iBlkX;
        iSubBlkY[0] = iBlkY;
        iSubBlkX[1] = iBdyX;
        iSubBlkY[1] = iBlkY;
        iSubBlkX[2] = iBlkX;
        iSubBlkY[2] = iBdyY;
        iSubBlkX[3] = iBdyX;
        iSubBlkY[3] = iBdyY;
        if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1])&& 
           sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[2], iSubBlkY[2])&&
           sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[3], iSubBlkY[3]))
        { //2Nx2N;
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBlkWidth, iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
        else if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1])&& 
                sameMotion(pCurrPic, iSubBlkX[2], iSubBlkY[2], iSubBlkX[3], iSubBlkY[3]))
        { //2NxN;
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBlkWidth, iBdyY - iBlkY);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

          cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
          cBlkSize.set(iBlkWidth, iBlkY + iBlkHeight - iSubBlkY[2]);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
        else if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[2], iSubBlkY[2])&& 
                sameMotion(pCurrPic, iSubBlkX[1], iSubBlkY[1], iSubBlkX[3], iSubBlkY[3]))
        { //Nx2N;
          cBlkPos.set(iBlkX, iBlkY);
          cBlkSize.set(iBdyX - iBlkX, iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      

          cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
          cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBlkHeight);
          calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
        }
        else
        {
          if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[1], iSubBlkY[1]))
          {//subblock0, subblock1;
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBlkWidth, iBdyY - iBlkY);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
            //subblock2;
            cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
            cBlkSize.set(iBdyX-iSubBlkX[2], iBlkY + iBlkHeight - iSubBlkY[2]);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
            //subblock3;
            cBlkPos.set(iSubBlkX[3], iSubBlkY[3]);
            cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[3], iBlkY + iBlkHeight - iSubBlkY[3]);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else if(sameMotion(pCurrPic, iSubBlkX[0], iSubBlkY[0], iSubBlkX[2], iSubBlkY[2]))
          {//subblock0, subblock2;
            cBlkPos.set(iBlkX, iBlkY);
            cBlkSize.set(iBdyX - iBlkX, iBlkHeight);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
            //subblock1;
            cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
            cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBdyY - iSubBlkY[1]);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
            //subblock3;
            cBlkPos.set(iSubBlkX[3], iSubBlkY[3]);
            cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[3], iBlkY + iBlkHeight - iSubBlkY[3]);
            calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
          }
          else
          {
              if(sameMotion(pCurrPic, iSubBlkX[1], iSubBlkY[1], iSubBlkX[3], iSubBlkY[3]))
              {
                //subblock0;
                cBlkPos.set(iBlkX, iBlkY);
                cBlkSize.set(iBdyX - iBlkX, iBdyY - iBlkY);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock2;
                cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
                cBlkSize.set(iBdyX-iSubBlkX[2], iBlkY + iBlkHeight - iSubBlkY[2]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock1/3;
                cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBlkHeight);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              }
              else if(sameMotion(pCurrPic, iSubBlkX[2], iSubBlkY[2], iSubBlkX[3], iSubBlkY[3]))
              {
                //subblock0;
                cBlkPos.set(iBlkX, iBlkY);
                cBlkSize.set(iBdyX - iBlkX, iBdyY - iBlkY);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock1;
                cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBdyY - iSubBlkY[1]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock2/3;
                cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
                cBlkSize.set(iBlkWidth, iBlkY + iBlkHeight - iSubBlkY[2]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              }
              else
              {
                //subblock0;
                cBlkPos.set(iBlkX, iBlkY);
                cBlkSize.set(iBdyX - iBlkX, iBdyY - iBlkY);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock1;
                cBlkPos.set(iSubBlkX[1], iSubBlkY[1]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[1], iBdyY - iSubBlkY[1]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock2;
                cBlkPos.set(iSubBlkX[2], iSubBlkY[2]);
                cBlkSize.set(iBdyX-iSubBlkX[2], iBlkY + iBlkHeight - iSubBlkY[2]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
                //subblock3;
                cBlkPos.set(iSubBlkX[3], iSubBlkY[3]);
                cBlkSize.set(iBlkX + iBlkWidth - iSubBlkX[3], iBlkY + iBlkHeight - iSubBlkY[3]);
                calcCompEnhILRBlock(pCurrPic, &cBlkPos, &cBlkSize, g_iSRNum, iDesignType);      
              }
          }
        }
      } //end partition check;

    }
  }
}
#endif
#endif

//! \}
